#ifndef PDCP_LOW_LEVEL_H
#define PDCP_LOW_LEVEL_H

#include <stdint.h>
#include "pdcp_names.h"


struct pdcp_device;
typedef struct pdcp_device pdcp_device;

struct pdcp_channel_link;
typedef struct pdcp_channel_link pdcp_channel_link;

struct pdcp_input_channel{
	uint8_t channel_index;
	uint16_t type;
	uint16_t profile;
	uint16_t terminal_profile;
	uint8_t transfer_enable;
	uint8_t required_input_channels;
	uint8_t desired_input_channels;
	uint16_t source_channel_node_id;
	uint16_t source_device_node_id;
	pdcp_device* owner;
	pdcp_channel_link* channel_link;
};
typedef struct pdcp_input_channel pdcp_input_channel;

struct pdcp_output_channel{
	uint8_t node_id;
	uint8_t channel_index;
	uint16_t type;
	uint16_t profile;
	uint8_t transfer_enable;
	uint8_t required_input_channels;
	uint8_t desired_input_channels;
	pdcp_device* owner;
	uint8_t n_channel_links;
	pdcp_channel_link** channel_links; // output channels can be connected to multiple input channels
	
};
typedef struct pdcp_output_channel pdcp_output_channel;

struct pdcp_channel_link{
	pdcp_input_channel* input_channel;
	pdcp_output_channel* output_channel;
	uint8_t link_status;
};

struct pdcp_device{
	char name[4];
	uint16_t vendor_id;
	uint16_t product_id;
	uint16_t serial_num;
	uint8_t node_id;
	uint16_t type;
	uint16_t profile;
	uint8_t n_input_channels;
	pdcp_input_channel** input_channels;
	uint8_t n_output_channels;
	pdcp_output_channel** output_channels;	
};


struct pdcp_system_info{
	uint8_t n_devices;
	uint8_t n_missing_devices;
	uint8_t n_input_channels;
	uint8_t n_output_channels;
	uint8_t n_channel_links;
	pdcp_device* devices;
	pdcp_device* missing_devices;
	pdcp_input_channel* input_channels;
	pdcp_output_channel* output_channels;
	pdcp_channel_link* channel_links;
};
typedef struct pdcp_system_info pdcp_system_info;

#define LINK_STATUS_ACTIVE 2
#define LINK_STATUS_INACTIVE 1

// Initializes low_level module and puts fills the out variable with all 
// info about devices, channels, and channel links on the system using explore_system().
return_code low_level_init(pdcp_system_info* out);

// frees arrays malloced by explore_system
void clean_up_system_info(pdcp_system_info* out);

// returns: (new == old)
int refresh_system_info(pdcp_system_info* out);

// Fills the out variable with all 
// info about devices, channels, and channel links on the system
// It mallocs quite a lot of arrays, so use clean_up_system_info to 
// dispose of system info variable.
return_code explore_system(pdcp_system_info* out);


// return_code get_system_info(pdcp_system_info* out);


// return_code explore_devices(pdcp_system_info* in_out);
// return_code explore_channels(pdcp_system_info* system_info);
// return_code explore_channel_links(pdcp_system_info* system_info);


// Puts value into parameter_value variable if parameter_size is large enough.
return_code get_device_parameter(const pdcp_device* dev, const uint8_t parameter_id, uint8_t* parameter_value, const uint16_t parameter_size);
return_code set_device_parameter(const pdcp_device* dev, const uint8_t parameter_id, const uint8_t* parameter_value, const uint16_t parameter_size);

// Puts value into parameter_value variable if parameter_size is large enough.
return_code get_input_channel_parameter(const pdcp_input_channel* chan, const uint8_t parameter_id, uint8_t* parameter_value, const uint16_t parameter_size);
// Puts value into parameter_value variable if parameter_size is large enough.
return_code get_output_channel_parameter(const pdcp_output_channel* chan, const uint8_t parameter_id, uint8_t* parameter_value, const uint16_t parameter_size);
return_code set_input_channel_parameter(const pdcp_input_channel* chan, const uint8_t parameter_id, const uint8_t* parameter_value, const uint16_t parameter_size);
return_code set_output_channel_parameter(const pdcp_output_channel* chan, const uint8_t parameter_id, const uint8_t* parameter_value, const uint16_t parameter_size);

// make_link_raw and make_channel_link both do the same, but with different arguments.
return_code make_channel_link_raw(pdcp_system_info* system_info, uint16_t out_vid, uint16_t out_pid, uint16_t out_snum, uint8_t out_channel_index, uint16_t in_vid, uint16_t in_pid, uint16_t in_snum, uint8_t in_channel_index);
return_code make_channel_link(pdcp_system_info* system_info, pdcp_input_channel* input_channel, pdcp_output_channel* output_channel);
return_code remove_channel_link(pdcp_system_info* system_info, pdcp_channel_link* link);
		
return_code passthrough_command(pdcp_device* dev, uint8_t data_length, uint8_t* can_data);

int links_equal(pdcp_channel_link*, pdcp_channel_link*);
int input_channels_equal(pdcp_input_channel*, pdcp_input_channel*);
int output_channels_equal(pdcp_output_channel* chan1, pdcp_output_channel* chan2);
int devices_equal(pdcp_device*, pdcp_device*);
int system_info_equal(pdcp_system_info* system1, pdcp_system_info* system2);

int is_missing(pdcp_system_info*, pdcp_device*);

pdcp_device* find_present_device(pdcp_system_info* system_info, uint16_t vid, uint16_t pid, uint16_t snum);
pdcp_device* find_missing_device(pdcp_system_info* system_info, uint16_t vid, uint16_t pid, uint16_t snum);
pdcp_input_channel* find_ichan_in_device(pdcp_device* device, uint8_t channel_index);
pdcp_output_channel* find_ochan_in_device(pdcp_device* device, uint8_t channel_index);


pdcp_input_channel*  find_or_add_ichan(pdcp_system_info* system_info, uint16_t in_vendor_id, 
					uint16_t in_product_id, uint16_t in_serial_num, 
					uint8_t in_channel_index, uint16_t type, uint16_t profile);
pdcp_output_channel* find_or_add_ochan(pdcp_system_info* system_info, uint16_t out_vendor_id, 
					uint16_t out_product_id, uint16_t out_serial_num, 
					uint8_t out_channel_index, uint16_t type, uint16_t profile);

// Prints only 5 lines (system_info.n_...)
void print_system_info_short(pdcp_system_info* system_info);
// Prints lots of info about all devices, channels, and channel links on the system.
void print_system_info(pdcp_system_info* system_info);


// void print_device(pdcp_system_info* system_info, pdcp_device* device);
// void print_input_channel(pdcp_system_info* system_info, pdcp_input_channel* channel);
// void print_output_channel(pdcp_system_info* system_info, pdcp_output_channel* channel);
// void print_channel_link(pdcp_system_info* system_info, pdcp_channel_link* link);



#endif
