#define _BSD_SOURCE

#include "pdcp_driver.h"
#include "pdcp_names.h"
#include "pdcp_malloc.h"
#include <stdio.h>
#include <stdint.h>
#include <errno.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>


static const uint8_t print_pos = 0;
static const uint8_t print_vars = 0;
static const uint8_t print_msgs = 0;


// dynamically allocates, call driver_free_msg_array() to free. 
uart_msg_array* driver_new_msg_array(const uint16_t payload_size);
void driver_free_msg_array(uart_msg_array* msg);

// when return_code is TOO_SMALL, make out with driver_new_msg_array(in->payload_size);
return_code driver_struct_to_array(const uart_msg_struct* in, uart_msg_array* out);

uint16_t driver_calculate_array_size(const uint16_t payload_size);

// does nothing at this time
uint8_t driver_get_crc8(const uart_msg_struct* in);
// does nothing at this time
uint16_t driver_get_crc16(const uart_msg_struct* in);


const uint16_t MIN_PDCP_UART_MSG_SIZE = 8;
const uint16_t PAYLOAD_OVERHEAD = 2;
const int MESSAGE_READY = -1;
const speed_t UART_SPEED = B57600;


/*uart_conn driver_conn;*/




uart_msg_array* driver_new_msg_array(const uint16_t payload_size){
	
	PRINT_POS(driver_new_msg_array);

	//printf("new_msg_array\n");
	uart_msg_array* ret;
	
	ret = (uart_msg_array*)pdcp_malloc(sizeof(uart_msg_array) + driver_calculate_array_size(payload_size)*sizeof(uint8_t));
	
	
	PRINTF_VAR("new array ptr: %x\n", (unsigned int)ret);
	
	
	if (ret != NULL) // pdcp_malloc succeeded
	{
		ret->array = (uint8_t*)ret+sizeof(uart_msg_array);
		ret->length = driver_calculate_array_size(payload_size);
	}
	
	//printf("new_msg_array done\n");
	return(ret);
}


void driver_free_msg_array(uart_msg_array* msg){
	PRINT_POS(driver_free_msg_array);
	PRINTF_VAR("freeing array ptr: %x\n", (unsigned int)msg);
	
	pdcp_free((void*)msg);
}

return_code driver_struct_to_array(const uart_msg_struct* in, uart_msg_array* out){
	PRINT_POS(driver_struct_to_array);
	
	if (out == NULL){
		return BAD_POINTER;
	}
	else if (out->length < driver_calculate_array_size(in->payload_size)){
		return TOO_SMALL;
	}
	else{
		out->array[0] = 0xFE;
		out->array[1] = 0x19;
		out->array[2] = in->msg_id & 0x00FF;
		out->array[3] = in->msg_id >> 8;
		out->array[4] = in->ref_id;
		out->array[5] = driver_get_crc8(in); // CRC8
		out->array[6] = (driver_calculate_array_size(in->payload_size)-8) & 0x00FF;
		out->array[7] = (driver_calculate_array_size(in->payload_size)-8) << 8;
		//printf("struct_to_array byte 8\n");
		if(in->payload_size != 0){
			memcpy(&out->array[MIN_PDCP_UART_MSG_SIZE], in->payload, in->payload_size*sizeof(uint8_t)); // payload
			out->array[MIN_PDCP_UART_MSG_SIZE + in->payload_size] = driver_get_crc16(in) & 0x00FF; // CRC16 LSB
			out->array[MIN_PDCP_UART_MSG_SIZE + in->payload_size + 1] = driver_get_crc16(in) >> 8; // CRC16 MSB
		}
		out->length = driver_calculate_array_size(in->payload_size);
	
		//printf("struct_to_array done\n");
		return OUT_VAR_FILLED;
	}
}

uint16_t driver_calculate_array_size(const uint16_t payload_size){
	//PRINT_POS(driver_calculate_array_size);
	uint16_t array_size;
	if (payload_size != 0) 
		array_size = (MIN_PDCP_UART_MSG_SIZE + payload_size + PAYLOAD_OVERHEAD);
	else 
		array_size = (MIN_PDCP_UART_MSG_SIZE);
		
	return (array_size);
}


int set_serial_attribs (uart_conn* conn, const speed_t speed, const int parity)
{
	// originally by user wallyk on stackoverflow.com (CC-BY-SA 2.5)
	// http://stackoverflow.com/questions/6947413/how-to-open-read-and-write-from-serial-port-in-c/6947758#6947758
	PRINT_POS(set_serial_attribs);
	struct termios tty;
	memset (&tty, 0, sizeof tty);
	if (tcgetattr (conn->read_fd, &tty) != 0)
	{
			printf ("error %d from tcgetattr (set_serial_attribs)\n", errno);
			return -1;
	}
	else if (tcgetattr (conn->write_fd, &tty) != 0)
	{
			printf ("error %d from tcgetattr (set_serial_attribs)\n", errno);
			return -1;
	}
	cfsetospeed (&tty, speed);
	cfsetispeed (&tty, speed);
	
	cfmakeraw(&tty);

	tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS8;     // 8-bit chars
	// disable IGNBRK for mismatched speed tests; otherwise receive break
	// as \000 chars
	tty.c_iflag &= ~IGNBRK;         // ignore break signal
	tty.c_lflag = 0;                // no signaling chars, no echo,
									// no canonical processing
	tty.c_oflag = 0;                // no remapping, no delays
	tty.c_cc[VMIN]  = 0;            // read doesn't block
	tty.c_cc[VTIME] = 5;            // 0.5 seconds read timeout

	tty.c_iflag &= ~(IXON | IXOFF | IXANY); // shut off xon/xoff ctrl

	tty.c_cflag |= (CLOCAL | CREAD);// ignore modem controls,
									// enable reading
	tty.c_cflag &= ~(PARENB | PARODD);      // shut off parity
	tty.c_cflag |= parity;
	tty.c_cflag &= ~CSTOPB;
	//tty.c_cflag &= ~CRTSCTS;

	if (tcsetattr (conn->read_fd, TCSANOW, &tty) != 0)
	{
			printf ("error %d from tcsetattr setting attributes to read_fd\n", errno);
			return -1;
	}
	else if (tcsetattr (conn->write_fd, TCSANOW, &tty) != 0)
	{
			printf ("error %d from tcsetattr setting attributes to write_fd\n", errno);
			return -1;
	}
	return 0;
}

void set_blocking (uart_conn* conn, const int should_block)
{
	// originally by user wallyk on stackoverflow.com (CC-BY-SA 2.5)
	// http://stackoverflow.com/questions/6947413/how-to-open-read-and-write-from-serial-port-in-c/6947758#6947758
	PRINT_POS(set_blocking);
	struct termios tty;
	memset (&tty, 0, sizeof tty);
	if (tcgetattr (conn->read_fd, &tty) != 0)
	{
			printf ("error %d from tggetattr (set_blocking)\n", errno);
			return;
	}

	tty.c_cc[VMIN]  = should_block ? 1 : 0;
	tty.c_cc[VTIME] = 5;            // 0.5 seconds read timeout

	if (tcsetattr (conn->read_fd, TCSANOW, &tty) != 0)
			printf ("error %d setting term attributes\n", errno);
}


void driver_init(const char* read_portname, const char* write_portname, uart_conn* out_conn){
	
	PRINT_POS(driver_init);
	out_conn->read_fd = open (read_portname, O_RDWR | O_SYNC | O_NOCTTY);
	if (out_conn->read_fd < 0)
	{
		printf ("error opening %s\n", read_portname);
		return;
	}
	if (strcmp(read_portname, write_portname)){
		out_conn->write_fd = open (write_portname, O_RDWR | O_SYNC | O_NOCTTY);
		if (out_conn->write_fd < 0)
		{
			printf ("error opening %s\n", write_portname);
			return;
		}
	}
	else{
		out_conn->write_fd = out_conn->read_fd;
	}
	
	

	set_serial_attribs (out_conn, UART_SPEED, 0);  // set speed to defined speed, 8n1 (no parity)
	set_blocking (out_conn, 0);                // set no blocking
	
	
}

ssize_t driver_send_msg(uart_conn* conn, const uart_msg_struct* in){
	PRINT_POS(driver_send_msg);
	///printf("send_msg\n");
	uart_msg_array* msg_array = driver_new_msg_array(in->payload_size);
	if (msg_array == NULL){
		return -2;
	}
	
	return_code ret = driver_struct_to_array(in, msg_array);
	
	if (ret != OUT_VAR_FILLED){
		return -2;
	}
	
	
	//printf("send_msg write\n");
	ssize_t write_count = write(conn->write_fd, msg_array->array, msg_array->length * sizeof(uint8_t));
	
	driver_free_msg_array(msg_array);
	
	return write_count;
}



return_code driver_check_for_msg(uart_conn* conn, uart_msg_struct* out){
	PRINT_POS(driver_check_for_msg);
	uint8_t read_byte;
	uint16_t crc16;
	static int state = 0;
	//static int msg_ready = 0;
	static uart_msg_struct buffer = {
		.payload = NULL
	};
	//buffer.payload = (uint8_t*)pdcp_malloc(1); // to have something to realloc()
	//static int payload_size;
	int read_count = 0;
	
	if (out == NULL){
		return BAD_POINTER;
	}
	
	if (state!=MESSAGE_READY)
		read_count = read(conn->read_fd, &read_byte, 1);
	
	while (read_count > 0 && state != MESSAGE_READY){
		PRINT_VAR(state, d);
		switch(state){
			case 0:
				if(read_byte == 0xFE)
					state = 1;
				break;
			case 1:
				if(read_byte == 0x19)
					state = 2;
				else
					state = 0;
				break;
			case 2:
				buffer.msg_id = read_byte;
				state++;
				break;
			case 3:
				buffer.msg_id += (uint16_t)read_byte << 8;
				state++;
				break;
			case 4:
				buffer.ref_id = read_byte;
				if(buffer.ref_id != out->ref_id)
					state = 0;
				else 
					state++;
				break;
			case 5:
				state++;
				break;
			case 6:
				buffer.payload_size = (read_byte == 0) ? read_byte : read_byte-2;
				
				state++;
				break;
			case 7:
				buffer.payload_size += (uint16_t)read_byte << 8;
				state++;
				if (buffer.payload_size == 0){
					state = MESSAGE_READY;
				}
				else{
					buffer.payload = (uint8_t*)pdcp_malloc(buffer.payload_size * sizeof(uint8_t));
				}
				break;
			default: // state >= 8:				
				if(state - 8 == buffer.payload_size){
					PRINTF_MSG("crc16 1st byte\n");
					crc16 = read_byte;
					state++;
				}
				else if(state - 8 == (buffer.payload_size + 1)){
					
					PRINTF_MSG("crc16 2nd byte\n");
					
					crc16 += read_byte << 8;
					state = MESSAGE_READY;
				}
				else{
					
					PRINTF_MSG("payload %d\n", state-8);
					
					buffer.payload[state-8] = read_byte;
					state++;
				}
				break;
				
		}
		
		if (state!=MESSAGE_READY)
			read_count = read(conn->read_fd, &read_byte, 1);
			
			
	}
	
	PRINT_POS_AUTO();
	
	if (state == MESSAGE_READY){
		if (out->payload_size < buffer.payload_size){
			conn->payload_size_ready = buffer.payload_size;
			conn->ref_id_ready = buffer.ref_id;
			conn->message_is_ready = 1;
			PRINT_POS_AUTO();
			return TOO_SMALL;
		
		}
		else{
			conn->message_is_ready = 0;
			state = 0;
			out->msg_id = buffer.msg_id;
			out->ref_id = buffer.ref_id;
			out->payload_size = buffer.payload_size;
			memcpy((void*)(out->payload), (void*)(buffer.payload), buffer.payload_size*sizeof(uint8_t));
			pdcp_free(buffer.payload);
			buffer.payload = NULL;
			
			PRINT_POS_AUTO();
			return OUT_VAR_FILLED;

		}
	}

	PRINT_POS_AUTO();
	return NO_ACTION;
	
}

return_code driver_blocking_read(uart_conn* conn, uart_msg_struct* out){
	PRINT_POS(driver_blocking_read);
	
	// uint8_t* dummy_payload = NULL;
	// uint8_t* payload_backup = NULL;
	
	// int done = 0;
	
	int loop_count = 0;
	int timeout = 500;
	
	return_code ret = NO_ACTION;
	do{
		ret = driver_check_for_msg(conn, out);
		
		// if ((ret == BAD_POINTER || ret == TOO_SMALL) && conn->ref_id_ready != ref_id){
			// payload_backup = out->payload;
			// dummy_payload = pdcp_malloc(conn->payload_size_ready * sizeof(uint8_t));
			// out->payload = dummy_payload;
		// }
		// else if(ret == OUT_VAR_FILLED && dummy_payload != NULL){
			// out
		// }
		
		
		PRINTF_VAR("ret: %d\n", ret);
		
		loop_count++;
		
		if (loop_count == 10)
			timeout = 500000;
		else if (loop_count == 11)
			break;
		usleep(timeout);
	} while(ret == NO_ACTION);
	
	if  (ret != OUT_VAR_FILLED){
		printf("read failed: ret: %d\n", ret);
		//assert (ret == OUT_VAR_FILLED);
	}
	return ret;
}

uart_msg_struct* driver_new_msg_struct(uint16_t payload_size){
	PRINT_POS(driver_new_msg_struct);
	uart_msg_struct* new_struct;
	new_struct = (uart_msg_struct*) pdcp_malloc(sizeof(uart_msg_struct) + payload_size*sizeof(uint8_t));
	
	PRINTF_VAR("new struct ptr: %x\n", (unsigned int)new_struct);
	
	
	if(new_struct == NULL){
	}
	else if (payload_size == 0){
		new_struct->payload_size = payload_size;
		new_struct->payload = NULL;
	}
	else{
		new_struct->payload_size = payload_size;
		new_struct->payload = (uint8_t*)new_struct + sizeof(uart_msg_struct);//&new_struct[1];
	}

	return new_struct;
}


void driver_free_msg_struct(uart_msg_struct* msg){
	
	PRINT_POS(driver_free_msg_struct);
	PRINTF_VAR("freeing struct ptr: %x\n", (unsigned int)msg);
	
	pdcp_free((void*)msg);
}

uint8_t driver_get_crc8(const uart_msg_struct* in){
	return 0;
}

uint16_t driver_get_crc16(const uart_msg_struct* in){
	return 0;
}


