#define _BSD_SOURCE

#include "pdcp_low_level.h"
#include "pdcp_driver.h"
#include "pdcp_names.h"
#include "pdcp_malloc.h"
#include "pdcp_error.h"
#include <assert.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

return_code get_system_info(pdcp_system_info* out);
return_code explore_devices(pdcp_system_info* in_out);
return_code explore_channels(pdcp_system_info* system_info);
return_code explore_channel_links(pdcp_system_info* system_info);

return_code explore_device(const uint8_t device_num, pdcp_device* out);
return_code explore_input_channel(const pdcp_system_info* system_info, const uint8_t channel_num, pdcp_input_channel* out);
return_code explore_output_channel(const pdcp_system_info* system_info, const uint8_t channel_num, pdcp_output_channel* out);
return_code explore_channel_link(pdcp_system_info* system_info, const uint8_t link_num, pdcp_channel_link* out);

// search only among present devices
pdcp_device* find_present_device(pdcp_system_info* system_info, uint16_t vid, uint16_t pid, uint16_t snum);

// search only among missing devices
pdcp_device* find_missing_device(pdcp_system_info* system_info, uint16_t vid, uint16_t pid, uint16_t snum);

pdcp_input_channel* find_ichan_in_device(pdcp_device* device, uint8_t channel_index);
pdcp_output_channel* find_ochan_in_device(pdcp_device* device, uint8_t channel_index);

void print_device(pdcp_system_info* system_info, pdcp_device* device);
void print_input_channel(pdcp_system_info* system_info, pdcp_input_channel* channel);
void print_output_channel(pdcp_system_info* system_info, pdcp_output_channel* channel);
void print_channel_link(pdcp_system_info* system_info, pdcp_channel_link* link);

void initialize_pdcp_device(pdcp_device* device, uint16_t vendor_id, uint16_t product_id, uint16_t serial_num, uint8_t node_id, uint16_t type, uint16_t profile);
void initialize_pdcp_output_channel (pdcp_output_channel* channel, uint8_t node_id, uint8_t channel_index, uint16_t type, uint16_t profile, uint8_t transfer_enable, pdcp_device* owner);
void initialize_pdcp_input_channel (pdcp_input_channel* channel, uint8_t channel_index, uint16_t type, uint16_t profile, uint8_t transfer_enable, pdcp_device* owner);



#define GET_INFO(REQUEST_PAYLOAD_SIZE, REQUEST_PAYLOAD_CONTENTS, RESPONSE_PAYLOAD_SIZE, REQUEST_MSG_ID) \
	uint8_t request_payload[] = REQUEST_PAYLOAD_CONTENTS;\
	uint8_t ref_id = get_ref_id(); \
	uart_msg_struct request = { \
		.msg_id = REQUEST_MSG_ID, \
		.ref_id = ref_id, \
		.payload_size = REQUEST_PAYLOAD_SIZE, \
		.payload = request_payload \
	};\
	driver_send_msg(&conn, &request); \
	uint8_t response_payload[RESPONSE_PAYLOAD_SIZE];\
	uart_msg_struct response = { \
		.ref_id = ref_id, \
		.payload = response_payload, \
		.payload_size = RESPONSE_PAYLOAD_SIZE \
	};\
	return_code ret;\
	ret = driver_blocking_read(&conn, &response);\
/*	#ifdef TEST \*/
/*	if(ret == NO_ACTION){ assert(!"no reply"); }  Bus arbitrator not responding \*/
/*	#endif*/

	
#ifdef PRINT_POSITION
const uint8_t print_pos = 1;
#else
const uint8_t print_pos = 0;
#endif

#ifdef PRINT_VARS
const uint8_t print_vars = 1;
#else
const uint8_t print_vars = 0;
#endif

#ifdef PRINT_MSGS
const uint8_t print_msgs = 1;
#else
const uint8_t print_msgs = 0;
#endif

uint8_t get_ref_id();

uart_conn conn;
uint8_t ref_id;
//uint8_t last_error_code;

pdcp_device null_device = {
	.vendor_id = 0xFFFF,
	.product_id = 0xFFFF,
	.serial_num = 0xFFFF
};
pdcp_output_channel null_output_channel = {
	.owner = &null_device,
	.channel_index = 0xFF
};



return_code low_level_init(pdcp_system_info* out){
	PRINT_POS(low_level_init);
	
	PRINT_VAR(print_vars, d);
	PRINT_VAR(print_pos, d);
		
	if(out == NULL){
		PRINT_POS_AUTO();
		return BAD_POINTER;
	}
	
	driver_init(READ_PATH, WRITE_PATH, &conn);
	
	return_code ret = explore_system(out);
	
	PRINT_POS_AUTO();
	
	return ret;
}


return_code explore_system(pdcp_system_info* out){
	return_code ret = get_system_info(out);
	if (ret == ERROR_CONDITION){
		printf("get_system_info returned error\n");
		return ret;
	}
	ret = explore_devices(out);
	if (ret == ERROR_CONDITION){
		printf("get_devices returned error\n");
		return ret;
	}
	ret = explore_channels(out);
	if (ret == ERROR_CONDITION){
		printf("get_channels returned error\n");
		return ret;
	}
	ret = explore_channel_links(out);
	if (ret == ERROR_CONDITION){
		printf("get_channel_links returned error\n");
		return ret;
	}
	return OUT_VAR_FILLED;
}


void clean_up_system_info(pdcp_system_info* out){
	PRINT_POS(clean_up_system_info);
	for(int i = 0; i < out->n_devices; i++){
		pdcp_free(out->devices[i].input_channels);
		pdcp_free(out->devices[i].output_channels);
	}
	for(int i = 0; i < out->n_output_channels; i++){
		pdcp_free(out->output_channels[i].channel_links);
	}
	for(int i = 0; i < out->n_missing_devices; i++){
		for (int j = 0; j < out->missing_devices[i].n_input_channels; j++){
			pdcp_free(out->missing_devices[i].input_channels[j]);
		}
		for (int j = 0; j < out->missing_devices[i].n_output_channels; j++){
			//for(int k = 0; k < out->missing_devices[i].output_channels[j].n_channel_links; k++){
			pdcp_free(out->missing_devices[i].output_channels[j]->channel_links);
			//}
			pdcp_free(out->missing_devices[i].output_channels[j]);
		}
		pdcp_free(out->missing_devices[i].input_channels);
		pdcp_free(out->missing_devices[i].output_channels);
	}

	pdcp_free(out->devices);
	pdcp_free(out->input_channels);
	pdcp_free(out->output_channels);
	pdcp_free(out->channel_links);
	pdcp_free(out->missing_devices);
	

}

int refresh_system_info(pdcp_system_info* out){
	PRINT_POS_AUTO();
	
	pdcp_system_info new_system;
	return_code ret = explore_system(&new_system);
	
	if(ret == OUT_VAR_FILLED){
		int equal = system_info_equal(out, &new_system);
	
		clean_up_system_info(out);	
	
		memcpy(out, &new_system, sizeof(pdcp_system_info));
		
		return equal;
	}
	else{
		printf("explore system error\n");
		
		return 1;
	}
}

return_code get_system_info(pdcp_system_info* out){
	PRINT_POS(get_system_info);
	
	if (out == NULL){
		PRINT_POS_AUTO();
		return BAD_POINTER;
	}
	
	GET_INFO(0, {0}, 4, MSG_ID_GET_SYSTEM_INFO)

	
	
	PRINTF_VAR("conn.payload_size_ready: %d\n", conn.payload_size_ready);
	PRINTF_VAR("ret: %d\n", ret);
	
	#ifndef TEST
	assert(ret == OUT_VAR_FILLED);
	#endif
	assert(response.ref_id == request.ref_id);
	
	
	switch(response.msg_id){
		case MSG_ID_GET_SYSTEM_INFO_SUCCESS:
			out->n_devices = response.payload[0];
			out->n_input_channels = response.payload[1];
			out->n_output_channels = response.payload[2];
			out->n_channel_links = response.payload[3];
			out->n_missing_devices = 0;
			out->devices = NULL;//(pdcp_device*)pdcp_malloc(out->n_devices * sizeof(pdcp_device));
			out->input_channels = NULL;
			out->output_channels = NULL;
			out->channel_links = NULL;
			out->missing_devices = pdcp_malloc(256 * sizeof(pdcp_device));
			break;
			
		case MSG_ID_GET_SYSTEM_INFO_FAILURE:
			pdcp_errno = response.payload[0];
			return ERROR_CONDITION;
			break;
			
		default:
			
			printf("msg_id = %X\n", response.msg_id);
			
			assert(!"expected system info msg id");
			break;
	
	}
	return OUT_VAR_FILLED;
}

return_code explore_device(const uint8_t device_num, pdcp_device* out){
	PRINT_POS(explore_device);
	
	if(out == NULL){
		return BAD_POINTER;
	}
	
	GET_INFO(1,{device_num}, 11, MSG_ID_GET_DEVICE_INFO)
	
	
	 
	PRINT_VAR(conn.payload_size_ready, d);
	
	assert(ret == OUT_VAR_FILLED);
	assert(response.ref_id == request.ref_id);

	switch(response.msg_id){
		case MSG_ID_GET_DEVICE_INFO_SUCCESS:
//			strcpy(out->name, "D");
			out->name[0] = 'D';
			out->name[1] = '\0';
			out->vendor_id = response.payload[0] + (response.payload[1] << 8);
			out->product_id = response.payload[2] + (response.payload[3] << 8);
			out->serial_num = response.payload[4] + (response.payload[5] << 8);
			out->type = response.payload[7] + (response.payload[8] << 8);
			out->profile = response.payload[9] + (response.payload[10] << 8);
			out->n_input_channels = 0;
			out->n_output_channels = 0;
			out->input_channels = NULL;
			out->output_channels = NULL;
			
/*			#ifndef TEST*/
/*			uint8_t node_id;*/
/*			return_code ret = get_device_parameter(out, PARAM_ID_DEV_NODE_ID, &node_id, 1);*/
/*			check_return_code(ret, __LINE__);*/
/*			#endif*/
/*			*/
/*			out->node_id = node_id;*/
			
			break;
			
		case MSG_ID_GET_DEVICE_INFO_FAILURE:
			pdcp_errno = response.payload[0];
			return ERROR_CONDITION;
			break;
			
		default:
			 
			PRINTF_VAR("msg_id = %d\n", response.msg_id);
			
			assert(0);
			return NO_ACTION;
			break;
	
	}
	
	return OUT_VAR_FILLED;
}

return_code explore_devices(pdcp_system_info* in_out){
	PRINT_POS(explore_devices);
	
	if(in_out == NULL){
		return BAD_POINTER;
	}
	
	if (in_out->devices == NULL && in_out->n_devices > 0){
		in_out->devices = pdcp_malloc(in_out->n_devices * sizeof(pdcp_device));
	}
	
	for (int i = 0; i < in_out->n_devices; i++){
		return_code ret = explore_device(i+1, &in_out->devices[i]);
		
		if (ret == ERROR_CONDITION){
			return ret;
		}
	}
	
	return OUT_VAR_FILLED;
}

return_code explore_input_channel(const pdcp_system_info* system_info, const uint8_t channel_num, pdcp_input_channel* out){
	PRINT_POS(explore_input_channel);
	
	if(out == NULL){
		return BAD_POINTER;
	}
	
	GET_INFO(1, {channel_num}, 11, MSG_ID_GET_INPUT_CHANNEL_INFO)
	
	 
	PRINT_VAR(conn.payload_size_ready, d);
	
	assert(ret == OUT_VAR_FILLED);
	assert(response.ref_id == request.ref_id);
	
	switch(response.msg_id){
		uint16_t vendor_id;
		uint16_t product_id;
		uint16_t serial_num;
		case MSG_ID_GET_INPUT_CHANNEL_INFO_SUCCESS:
			vendor_id = response.payload[0] + (response.payload[1] << 8);
			product_id = response.payload[2] + (response.payload[3] << 8);
			serial_num = response.payload[4] + (response.payload[5] << 8);
			
			PRINTF_VAR("vid: %X, pid: %X, sn: %X\n", vendor_id, product_id, serial_num);
			
			
			out->owner = NULL;
			
			// find owner
			for (int i = 0; i < system_info->n_devices && out->owner == NULL; i++){
				pdcp_device* cur_device = &system_info->devices[i];
				
				 
				PRINTF_VAR("vid: %X, pid: %X, sn: %X\n", cur_device->vendor_id, cur_device->product_id, cur_device->serial_num);
				
				
				// if cur_device is the owner of the channel
				if (cur_device->vendor_id == vendor_id && cur_device->product_id == product_id  && cur_device->serial_num == serial_num){
					 
					PRINTF_MSG("found device\n");
				
					out->owner = cur_device;
					cur_device->n_input_channels++;
				}
			}
			
			if (out->owner == NULL){ // owner not found
			}
			else{
				uint16_t type = response.payload[7] + (response.payload[8] << 8);
				uint16_t profile = response.payload[9] + (response.payload[10] << 8);
				uint16_t channel_index = response.payload[6];
				
				out->channel_index = channel_index;
				
				uint8_t transfer_enable = 99;
				uint8_t required_input_channels;
				uint8_t desired_input_channels = 0;
				uint16_t terminal_profile;
				uint8_t source_node_ids[2] = {0,0};
				#ifndef TEST
				uint8_t terminal_profile_array[2];
				return_code ret = get_input_channel_parameter(out, PARAM_ID_CHAN_TERMINAL_PROFILE, &terminal_profile_array[0], 2);
				check_return_code(ret, __LINE__);
				terminal_profile = terminal_profile_array[0] + (terminal_profile_array[1] << 8);
				
/*				ret = get_input_channel_parameter(out, PARAM_ID_CHAN_TRANSFER_ENABLE, &transfer_enable, 1);*/
/*				check_return_code(ret, __LINE__);*/
				
				ret = get_input_channel_parameter(out, PARAM_ID_CHAN_REQUIRED_ICHANS, &required_input_channels, 1);
				check_return_code(ret, __LINE__);
				
/*				ret = get_input_channel_parameter(out, PARAM_ID_CHAN_DESIRED_ICHANS, &desired_input_channels, 1);*/
/*				check_return_code(ret, __LINE__);*/
				
/*				ret = get_input_channel_parameter(out, PARAM_ID_CHAN_SOURCE_NODE_IDS, &source_node_ids[0], 2);*/
/*				check_return_code(ret, __LINE__);*/
				
				#endif
				
				out->type = type;
				out->profile = profile;
				out->channel_link = NULL;
				out->transfer_enable = transfer_enable;
				out->terminal_profile = terminal_profile;
				out->required_input_channels = required_input_channels;
				out->desired_input_channels = desired_input_channels;
				out->source_channel_node_id = source_node_ids[0];
				out->source_device_node_id = source_node_ids[1];
			}
			break;
			
		case MSG_ID_GET_INPUT_CHANNEL_INFO_FAILURE:
			pdcp_errno = response.payload[0];
			return ERROR_CONDITION;
			break;
			
		default:
			 
			PRINTF_VAR("msg_id = %X\n", response.msg_id);
			
			
			assert(0);
			break;
	}
	
	return OUT_VAR_FILLED;
}


return_code explore_output_channel(const pdcp_system_info* system_info, const uint8_t channel_num, pdcp_output_channel* out){
	PRINT_POS(explore_output_channel);
	
	
	if(out == NULL){
		PRINT_POS_AUTO();
		return BAD_POINTER;
	}
	
	GET_INFO(1, {channel_num}, 11, MSG_ID_GET_OUTPUT_CHANNEL_INFO)
	
	 
	PRINT_VAR(conn.payload_size_ready, d);
	
	assert(ret == OUT_VAR_FILLED);
	assert(response.ref_id == request.ref_id);
	
	switch(response.msg_id){
		uint16_t vendor_id;
		uint16_t product_id;
		uint16_t serial_num;
		case MSG_ID_GET_OUTPUT_CHANNEL_INFO_SUCCESS:
			vendor_id = response.payload[0] + (response.payload[1] << 8);
			product_id = response.payload[2] + (response.payload[3] << 8);
			serial_num = response.payload[4] + (response.payload[5] << 8);
			
			 
			PRINTF_VAR("vid: %X, pid: %X, sn: %X\n", vendor_id, product_id, serial_num);
			
			
			out->owner = NULL;
			
			// find owner
			for (int i = 0; i < system_info->n_devices && out->owner == NULL; i++){
				pdcp_device* cur_device = &system_info->devices[i];
				
				 
				PRINTF_VAR("vid: %X, pid: %X, sn: %X\n", cur_device->vendor_id, cur_device->product_id, cur_device->serial_num);
				
				
				// if cur_device is the owner of the channel
				if (cur_device->vendor_id == vendor_id && cur_device->product_id == product_id  && cur_device->serial_num == serial_num){
					
					PRINTF_MSG("found device\n");
					
					out->owner = cur_device;
					cur_device->n_output_channels++;
				}
			}
			
			if (out->owner == NULL){ // owner not found
			
			}
			else{
				uint16_t type = response.payload[7] + (response.payload[8] << 8);
				uint16_t profile = response.payload[9] + (response.payload[10] << 8);
				uint16_t channel_index = response.payload[6];
				
				out->channel_index = channel_index;
				
				uint8_t node_id = 0;
				uint8_t transfer_enable = 99;
				uint8_t required_input_channels;
				uint8_t desired_input_channels = 0;
				#ifndef TEST
/*				return_code ret = get_output_channel_parameter(out, PARAM_ID_CHAN_NODE_ID, &node_id, 1);*/
/*				check_return_code(ret, __LINE__);*/
				
/*				ret = get_output_channel_parameter(out, PARAM_ID_CHAN_TRANSFER_ENABLE, &transfer_enable, 1);*/
/*				check_return_code(ret, __LINE__);*/
				
			 	ret = get_output_channel_parameter(out, PARAM_ID_CHAN_REQUIRED_ICHANS, &required_input_channels, 1);
				check_return_code(ret, __LINE__);
				
/*				ret = get_output_channel_parameter(out, PARAM_ID_CHAN_DESIRED_ICHANS, &desired_input_channels, 1);*/
/*				check_return_code(ret, __LINE__);*/
				#endif
				
				out->type = type;
				out->profile = profile;
				out->n_channel_links = 0;
				out->channel_links = NULL;
				out->transfer_enable = transfer_enable;
				out->required_input_channels = required_input_channels;
				out->desired_input_channels = desired_input_channels;
				out->node_id = node_id;
				
				

			}
			break;
			
		case MSG_ID_GET_OUTPUT_CHANNEL_INFO_FAILURE:
			pdcp_errno = response.payload[0];
			return ERROR_CONDITION;	
			break;
			
		default:
			
			PRINTF_VAR("msg_id = %X\n", response.msg_id);
			
			assert(0);
			break;
	}
	
	PRINT_POS(explore_output_channel done);
	return OUT_VAR_FILLED;
}


return_code explore_channels(pdcp_system_info* system_info){
	PRINT_POS(explore_channels);
	
	if(system_info == NULL){
		return BAD_POINTER;
	
	}
	
	return_code ret;
	
	if (system_info->input_channels == NULL && system_info->output_channels == NULL){
		
		PRINTF_VAR("system_info->n_input_channels: %X\n", system_info->n_input_channels);
		PRINTF_VAR("system_info->n_output_channels: %X\n", system_info->n_output_channels);
		
		system_info->input_channels = (pdcp_input_channel*)pdcp_malloc(system_info->n_input_channels * sizeof(pdcp_input_channel));
		system_info->output_channels = (pdcp_output_channel*)pdcp_malloc(system_info->n_output_channels * sizeof(pdcp_output_channel));		
	}
	else{ // something is wrong
		printf("explore_channels error: channel arrays not null\n");
		assert(0);
	}
	
	PRINTF_MSG("explore input\n");
	
	// explore input channels
	for(int i = 0; i < system_info->n_input_channels; i++){
		ret = explore_input_channel(system_info, i+1, &system_info->input_channels[i]);
		
		if (ret == ERROR_CONDITION){
			pdcp_free(system_info->input_channels);
			pdcp_free(system_info->output_channels);
			system_info->input_channels = NULL;
			system_info->output_channels = NULL;
			return ret;
		}
	}
	
	PRINTF_MSG("explore output\n");
	
	// explore output channels
	for(int i = 0; i < system_info->n_output_channels; i++){
		ret = explore_output_channel(system_info, i+1, &system_info->output_channels[i]);
		
		if (ret == ERROR_CONDITION){
			pdcp_free(system_info->input_channels);
			pdcp_free(system_info->output_channels);
			system_info->input_channels = NULL;
			system_info->output_channels = NULL;
			return ret;
		}
	}
	
	PRINTF_MSG("allocate in devices\n");
	
	// allocate channel arrays in devices
	for(int i = 0; i < system_info->n_devices; i++){
		pdcp_device* device = &system_info->devices[i];
		
		
		PRINTF_VAR("device->n_input_channels: %X\n", device->n_input_channels);
		PRINTF_VAR("device->n_output_channels: %X\n", device->n_output_channels);
		
		
		device->input_channels = (pdcp_input_channel**)pdcp_malloc(device->n_input_channels * sizeof(pdcp_input_channel*));
		device->output_channels = (pdcp_output_channel**)pdcp_malloc(device->n_output_channels * sizeof(pdcp_output_channel*));
		
		memset(device->input_channels, 0, device->n_input_channels * sizeof(pdcp_input_channel*));
		memset(device->output_channels, 0, device->n_output_channels * sizeof(pdcp_output_channel*));
	}
	
	PRINTF_MSG("fill input in devices\n");
	
	// fill input channel arrays in devices
	for(int i = 0; i < system_info->n_input_channels; i++){
		pdcp_input_channel* channel = &system_info->input_channels[i];
		
		
		PRINTF_VAR("owner: %X\n", (unsigned)channel->owner);
		

		int channel_num = 0;
		for (; channel->owner->input_channels[channel_num] != NULL; channel_num++){	}
		
		PRINTF_VAR("channel_num: %d\n", channel_num);
		
		channel->owner->input_channels[channel_num] = channel;
	}
	
	PRINTF_MSG("fill output in devices\n");
	
	// fill output channel arrays in devices
	for(int i = 0; i < system_info->n_output_channels; i++){
		pdcp_output_channel* channel = &system_info->output_channels[i];
		
		int channel_num = 0;
		for (; channel->owner->output_channels[channel_num] != NULL; channel_num++){}
		channel->owner->output_channels[channel_num] = channel;
	}
	
	PRINTF_MSG("explore_channels() done\n");
	
	return OUT_VAR_FILLED;
}



	
return_code explore_channel_link(pdcp_system_info* system_info, const uint8_t link_num, pdcp_channel_link* out){
	PRINT_POS(explore_channel_link);


	if(out == NULL){
		PRINT_POS_AUTO();
		return BAD_POINTER;
	}

	GET_INFO(1, {link_num}, 25, MSG_ID_GET_CHANNEL_LINK_INFO)
	
	
	PRINT_VAR(conn.payload_size_ready, d);
	
	assert(ret == OUT_VAR_FILLED);
	assert(response.ref_id == request.ref_id);
	
	switch(response.msg_id){
		uint16_t in_vendor_id;
		uint16_t in_product_id;
		uint16_t in_serial_num;
		uint8_t in_channel_index;
		uint16_t out_vendor_id;
		uint16_t out_product_id;
		uint16_t out_serial_num;
		uint8_t out_channel_index;
		uint8_t link_status;
		case MSG_ID_GET_CHANNEL_LINK_INFO_SUCCESS:
			out_vendor_id = response.payload[0] + (response.payload[1] << 8);
			out_product_id = response.payload[2] + (response.payload[3] << 8);
			out_serial_num = response.payload[4] + (response.payload[5] << 8);
			out_channel_index = response.payload[6];
			
			in_vendor_id = response.payload[7] + (response.payload[8] << 8);
			in_product_id = response.payload[9] + (response.payload[10] << 8);
			in_serial_num =  response.payload[11] + (response.payload[12] << 8);
			in_channel_index = response.payload[13];
			
			link_status = response.payload[14];

			
			out->input_channel = NULL;
			out->output_channel = NULL;
			
			
			// find input channel
			
			pdcp_device* in_owner = find_present_device(system_info, in_vendor_id, in_product_id, in_serial_num);
			if (in_owner == NULL){ // device missing
				
				in_owner = find_missing_device(system_info, in_vendor_id, in_product_id, in_serial_num);
				if (in_owner == NULL){ // device not yet registered as missing
					//RESIZE_ARRAY(system_info->missing_devices, system_info->n_missing_devices, system_info->n_missing_devices + 1, pdcp_device);
					system_info->n_missing_devices++;
					
					
					initialize_pdcp_device(&system_info->missing_devices[(system_info->n_missing_devices)-1], in_vendor_id, in_product_id, in_serial_num, 0xFF, 0, 0);
					
					in_owner = &system_info->missing_devices[system_info->n_missing_devices-1];
				}
				
				
				out->input_channel = find_ichan_in_device(in_owner, in_channel_index);
				
				if(out->input_channel == NULL){ // channel not registered yet
					RESIZE_ARRAY(in_owner->input_channels, in_owner->n_input_channels, in_owner->n_input_channels+1, pdcp_input_channel*);
					in_owner->n_input_channels++;
					in_owner->input_channels[in_owner->n_input_channels-1] = pdcp_malloc(sizeof(pdcp_input_channel));
										
					initialize_pdcp_input_channel(in_owner->input_channels[in_owner->n_input_channels-1], in_channel_index, 0, 0, 0, in_owner);
					
				}
			}
		
			out->input_channel = find_ichan_in_device(in_owner, in_channel_index);
			assert(out->input_channel != NULL);
			
			
			// find output channel
			
			pdcp_device* out_owner = find_present_device(system_info, out_vendor_id, out_product_id, out_serial_num);
			if (out_owner == NULL){ // device missing
				
				out_owner = find_missing_device(system_info, out_vendor_id, out_product_id, out_serial_num);
				if (out_owner == NULL){ // device not yet registered as missing
					//RESIZE_ARRAY(system_info->missing_devices, system_info->n_missing_devices, system_info->n_missing_devices + 1, pdcp_device);
					system_info->n_missing_devices++;
					
					printf("new device: %X\n", (unsigned)&system_info->missing_devices[(system_info->n_missing_devices-1)]);
					
					initialize_pdcp_device(&system_info->missing_devices[(system_info->n_missing_devices-1)], out_vendor_id, out_product_id, out_serial_num, 0xFF, 0, 0);
					
					out_owner = &system_info->missing_devices[system_info->n_missing_devices-1];
				}
				
				
				out->output_channel = find_ochan_in_device(out_owner, out_channel_index);
				
				if(out->output_channel == NULL){  // channel not registered yet
					RESIZE_ARRAY(out_owner->output_channels, out_owner->n_output_channels, out_owner->n_output_channels+1, pdcp_output_channel*);
					out_owner->n_output_channels++;
					out_owner->output_channels[out_owner->n_output_channels-1] = pdcp_malloc(sizeof(pdcp_output_channel));
					
					uint8_t type_profile[4];
					return_code ret = get_input_channel_parameter(out->input_channel, PARAM_ID_CHAN_SOURCE_TYPE_PROFILE, &type_profile[0], 4);
					check_return_code(ret, __LINE__);
					
					printf("new output channel: %X\n", (unsigned)out_owner->output_channels[out_owner->n_output_channels-1]);
					initialize_pdcp_output_channel(out_owner->output_channels[out_owner->n_output_channels-1], 0xFF, out_channel_index, type_profile[0]+(type_profile[1]<<8), type_profile[2]+(type_profile[3]<<8), 0, out_owner);
				}
			}
			
			out->output_channel = find_ochan_in_device(out_owner, out_channel_index);
			assert(out->output_channel != NULL);
			
			//printf("output channel: %X\n", out->output_channel);
			
			out->output_channel->n_channel_links++;
			
			//printf("output channel n_channel_links: %d\n", out->output_channel->n_channel_links);
			
			//printf("link: %X\n", out);
			
			
			out->link_status = link_status;

			break;
			
		case MSG_ID_GET_CHANNEL_LINK_INFO_FAILURE:
			pdcp_errno = response.payload[0];
			return ERROR_CONDITION;		
			break;
			
		default:
			printf("msg_id = %X\n", response.msg_id);
			assert(0);
			break;
	}
	
	return OUT_VAR_FILLED;
}


return_code explore_channel_links(pdcp_system_info* system_info){
	PRINT_POS(explore_channel_links);
	
	if(system_info == NULL){
		PRINT_POS_AUTO();
		return BAD_POINTER;
	}
	
	PRINT_VAR(system_info->n_channel_links, d);
	
	system_info->channel_links = (pdcp_channel_link*)pdcp_malloc(system_info->n_channel_links * sizeof(pdcp_channel_link));
	if (system_info->channel_links == NULL){
		return -1;
	}
	
/*	pdcp_device* missing_devices = NULL;*/
	

	
	for (int i = 0; i < system_info->n_channel_links; i++){	
		
		// printf("system_info->n_channel_links: %d\n", system_info->n_channel_links);
		
		// printf("link #%d\n", i);
		
		pdcp_channel_link* link = &system_info->channel_links[i];
	
		return_code ret = explore_channel_link(system_info, i+1, link);
		
		PRINT_POS_AUTO();
		
		PRINT_VAR(ret, d);
		
		if (ret == ERROR_CONDITION){
			return ret;
		}
	}
	
	for (int i = 0; i < system_info->n_channel_links; i++){	
		pdcp_channel_link* link = &system_info->channel_links[i];
		pdcp_input_channel* inchan = link->input_channel;
		pdcp_output_channel* outchan = link->output_channel;
		
		//printf("link: %X\n", link);
			
		// for output channel
		int index_in_array = 0;
		if (outchan->channel_links == NULL){ // allocate array
			outchan->channel_links = (pdcp_channel_link**)pdcp_malloc(outchan->n_channel_links * sizeof(pdcp_channel_link*));
			memset(outchan->channel_links, 0, outchan->n_channel_links * sizeof(pdcp_channel_link*));
		}
		else{// find vacant place in array
			for ( ; outchan->channel_links[index_in_array] != NULL; index_in_array++){}
		}
		
		PRINTF_VAR("link: %X\n", (unsigned)link);
		
		outchan->channel_links[index_in_array] = link;
		
		PRINTF_VAR("outchan->channel_links: %X\n", (unsigned)outchan->channel_links);
		PRINTF_VAR("inchan: %X\n", (unsigned)inchan);
		PRINTF_VAR("index_in_array: %d\n", index_in_array);
		
		inchan->channel_link = link;
	}
	return OUT_VAR_FILLED;
}


return_code get_parameter(const pdcp_device* dev, const uint8_t parameter_id, uint8_t* parameter_value, const uint16_t parameter_size, const uint8_t channel_index){
	
	
	uint8_t request_payload[] = {dev->vendor_id & 0xFF, dev->vendor_id >> 8, dev->product_id & 0xFF, dev->product_id >> 8, dev->serial_num & 0xFF, dev->serial_num >> 8, channel_index, parameter_id};
	uint8_t ref_id = get_ref_id();
	uart_msg_struct request = { 
		.msg_id = MSG_ID_GET_DEVICE_PARAMETER, 
		.ref_id = ref_id, 
		.payload_size = 8, 
		.payload = request_payload 
	};
	driver_send_msg(&conn, &request); 
	
	uart_msg_struct response = { 
		.ref_id = ref_id, 
		.payload = parameter_value, 
		.payload_size = parameter_size 
	};
	
	return_code ret = driver_blocking_read(&conn, &response);
	
	if (ret == OUT_VAR_FILLED){
		if (response.msg_id == MSG_ID_GET_DEVICE_PARAMETER_FAILURE){
			ret = ERROR_CONDITION;
			pdcp_errno = response.payload[0];
		}
		else{
			assert(response.msg_id == MSG_ID_GET_DEVICE_PARAMETER_SUCCESS);
		}
	}
	
	return ret;
	
}


return_code set_parameter(const pdcp_device* dev, const uint8_t parameter_id, const uint8_t* parameter_value, const uint16_t parameter_size, const uint8_t channel_index){
	PRINT_POS(set_parameter);
	uart_msg_struct* request = driver_new_msg_struct(8 + parameter_size);
			
	request->msg_id = MSG_ID_SET_DEVICE_PARAMETER;
	request->ref_id = get_ref_id();
	request->payload[0] = dev->vendor_id & 0xFF;
	request->payload[1] = dev->vendor_id >> 8;
	request->payload[2] = dev->product_id & 0xFF;
	request->payload[3] = dev->product_id >> 8; 
	request->payload[4] = dev->serial_num & 0xFF;
	request->payload[5] = dev->serial_num >> 8;
	request->payload[6] = channel_index;
	request->payload[7] = parameter_id;
	memcpy(&request->payload[8], parameter_value, parameter_size);

	
	PRINT_POS_AUTO();
	
	driver_send_msg(&conn, request);
	
	
	uint8_t response_payload;
	uart_msg_struct response = { 
		.ref_id = request->ref_id, 
		.payload = &response_payload, 
		.payload_size = 1 
	};
	
	PRINT_POS_AUTO();
	driver_free_msg_struct(request);
	
	PRINT_POS_AUTO();
	return_code ret = driver_blocking_read(&conn, &response);
	
	PRINT_POS_AUTO();
	if (ret == OUT_VAR_FILLED){
		if (response.msg_id == MSG_ID_SET_DEVICE_PARAMETER_FAILURE){
			ret = ERROR_CONDITION;
			
			pdcp_errno = response.payload[0];
		}
		else{
			assert(response.msg_id == MSG_ID_SET_DEVICE_PARAMETER_SUCCESS);
		}
	}
	
	PRINT_POS_AUTO();
	return ret;
}



return_code set_device_parameter(const pdcp_device* dev, const uint8_t parameter_id, const uint8_t* parameter_value, const uint16_t parameter_size){
	PRINT_POS_AUTO();
	return set_parameter(dev, parameter_id, parameter_value, parameter_size, 0);
}
return_code get_device_parameter(const pdcp_device* dev, const uint8_t parameter_id, uint8_t* parameter_value, const uint16_t parameter_size){
	PRINT_POS_AUTO();
	return get_parameter(dev, parameter_id, parameter_value, parameter_size, 0);
}
return_code get_input_channel_parameter(const pdcp_input_channel* chan, const uint8_t parameter_id, uint8_t* parameter_value, const uint16_t parameter_size){
	PRINT_POS_AUTO();
	return get_parameter (chan->owner, parameter_id, parameter_value, parameter_size, chan->channel_index);
}
return_code get_output_channel_parameter(const pdcp_output_channel* chan, const uint8_t parameter_id, uint8_t* parameter_value, const uint16_t parameter_size){
	PRINT_POS_AUTO();
	return get_parameter (chan->owner, parameter_id, parameter_value, parameter_size, chan->channel_index);
}
return_code set_input_channel_parameter(const pdcp_input_channel* chan, const uint8_t parameter_id, const uint8_t* parameter_value, const uint16_t parameter_size){
	PRINT_POS_AUTO();
	return set_parameter (chan->owner, parameter_id, parameter_value, parameter_size, chan->channel_index);
}
return_code set_output_channel_parameter(const pdcp_output_channel* chan, const uint8_t parameter_id, const uint8_t* parameter_value, const uint16_t parameter_size){
	PRINT_POS_AUTO();
	return set_parameter (chan->owner, parameter_id, parameter_value, parameter_size, chan->channel_index);
}




uint8_t get_ref_id(){

#ifdef TEST
	return 13;
	
#else
	static uint8_t ref_id = 0;
	return ref_id++;
#endif
	
}

return_code make_channel_link_raw(pdcp_system_info* system_info, uint16_t out_vid, uint16_t out_pid, uint16_t out_snum, uint8_t out_channel_index, uint16_t in_vid, uint16_t in_pid, uint16_t in_snum, uint8_t in_channel_index){
	uart_msg_struct request;
	request.msg_id = MSG_ID_CONFIGURE_CHANNEL_LINK;
	request.ref_id = get_ref_id();
	request.payload_size = 14;
	
	uint8_t payload[14];
	payload[0] = out_vid & 0xFF;
	payload[1] = out_vid >> 8;
	payload[2] = out_pid & 0xFF;
	payload[3] = out_pid >> 8;
	payload[4] = out_snum & 0xFF;
	payload[5] = out_snum >> 8;
	payload[6] = out_channel_index;
	payload[7] = in_vid & 0xFF;
	payload[8] = in_vid >> 8;
	payload[9] = in_pid & 0xFF;
	payload[10] = in_pid >> 8;
	payload[11] = in_snum & 0xFF;
	payload[12] = in_snum >> 8;
	payload[13] = in_channel_index;
	
	request.payload = &payload[0];
	
	driver_send_msg(&conn, &request);
	
	PRINT_POS_AUTO();
	
	uint8_t response_payload[20];
	uart_msg_struct response = {
		.ref_id = request.ref_id,
		.payload_size = 20,
		.payload = &response_payload[0]
	};
	
	return_code ret = driver_blocking_read(&conn, &response);
	
	
	switch (response.msg_id){
		case MSG_ID_CONFIGURE_CHANNEL_LINK_SUCCESS:
			PRINT_POS_AUTO();
			
			if(system_info != NULL){
				pdcp_device* idev = find_present_device(system_info, in_vid, in_pid, in_snum);
				pdcp_device* odev = find_present_device(system_info, out_vid, out_pid, out_snum);
				if (idev != NULL && odev != NULL){
					pdcp_input_channel* ichan = find_ichan_in_device(idev, in_channel_index);
					pdcp_output_channel* ochan = find_ochan_in_device(odev, out_channel_index);
					if (ichan != NULL && ochan != NULL){
						uint8_t type_profile[4];
						type_profile[0] = ochan->type & 0xFF;
						type_profile[1] = ochan->type >> 8;
						type_profile[2] = ochan->profile & 0xFF;
						type_profile[3] = ochan->profile >> 8;
						
						set_input_channel_parameter(ichan, PARAM_ID_CHAN_SOURCE_TYPE_PROFILE, &type_profile[0], 4);
					}
				}
			}
			
			break;
		case MSG_ID_CONFIGURE_CHANNEL_LINK_FAILURE:
			PRINT_POS_AUTO();
			
			pdcp_errno = response.payload[0];
			return ERROR_CONDITION;
			break;
		default:
			printf("msg_id: %X\n", response.msg_id);
			printf("ret: %d\n", ret);
			printf("size: %d\n", response.payload_size);
			printf("ref id: %d\n", response.ref_id);
			printf("request msg id: %d\n", request.msg_id);
			assert(!"wrong msg_id");
			break;
	}
	
	return OUT_VAR_FILLED;

}

return_code make_channel_link(pdcp_system_info* system_info, pdcp_input_channel* input_channel, pdcp_output_channel* output_channel){
	PRINT_POS_AUTO();
	
	if (system_info == NULL || input_channel == NULL || output_channel == NULL){
		return BAD_POINTER;
	}
	else if(input_channel->owner == NULL || output_channel->owner == NULL){
		return BAD_POINTER;
	}

	return_code ret = make_channel_link_raw(system_info,
						output_channel->owner->vendor_id, output_channel->owner->product_id, 
						output_channel->owner->serial_num, output_channel->channel_index, 
						input_channel->owner->vendor_id, input_channel->owner->product_id, 
						input_channel->owner->serial_num, input_channel->channel_index);

	//if (ret == OUT_VAR_FILLED) refresh_system_info(system_info);
	
	return ret;
}


return_code remove_channel_link(pdcp_system_info* system_info, pdcp_channel_link* link){
	return make_channel_link(system_info, link->input_channel, &null_output_channel);
}


return_code passthrough_command(pdcp_device* dev, uint8_t data_length, uint8_t* can_data){
	PRINT_POS_AUTO();
	
	if (dev == NULL || (can_data == NULL && data_length != 0) || data_length < 0 || data_length > 8){
		return BAD_POINTER;
	}
	
	uart_msg_struct request;
	request.msg_id = MSG_ID_PASSTHROUGH;
	request.ref_id = get_ref_id();
	request.payload_size = 15;
	
	uint8_t payload[15];
	payload[0] = dev->vendor_id & 0xFF;
	payload[1] = dev->vendor_id >> 8;
	payload[2] = dev->product_id & 0xFF;
	payload[3] = dev->product_id >> 8;
	payload[4] = dev->serial_num & 0xFF;
	payload[5] = dev->serial_num >> 8;
	payload[6] = data_length;
	for(int i = 0; i < data_length; i++){
		payload[7+i] = can_data[i]; 
	}
	for(int i = 7; i >= data_length; i--){
		payload[7+i] = 0; 
	}

	request.payload = &payload[0];
	
	driver_send_msg(&conn, &request);
	
	PRINT_POS_AUTO();
	
	uint8_t response_payload[15];
	uart_msg_struct response = {
		.ref_id = request.ref_id,
		.payload_size = 15,
		.payload = &response_payload[0]
	};
	
	return_code ret = driver_blocking_read(&conn, &response);
	
	
	switch (response.msg_id){
		case MSG_ID_PASSTHROUGH_SUCCESS:
			PRINT_POS_AUTO();
			
			break;
		case MSG_ID_PASSTHROUGH_FAILURE:
			PRINT_POS_AUTO();
			
			pdcp_errno = response.payload[0];
			return ERROR_CONDITION;
			break;
		default:
			printf("msg_id: %X\n", response.msg_id);
			printf("ret: %d\n", ret);
			assert(0);
			break;
	}
	
	return OUT_VAR_FILLED;
}


void initialize_pdcp_device(pdcp_device* device, uint16_t vendor_id, uint16_t product_id, uint16_t serial_num, uint8_t node_id, uint16_t type, uint16_t profile){
	device->vendor_id = vendor_id;
	device->product_id = product_id;
	device->serial_num = serial_num;
	device->node_id = node_id;
	device->type = type;
	device->profile = profile;
	device->n_input_channels = 0;
	device->input_channels = NULL;
	device->n_output_channels = 0;
	device->output_channels = NULL;
	
}

void initialize_pdcp_output_channel (pdcp_output_channel* channel, uint8_t node_id, uint8_t channel_index, uint16_t type, uint16_t profile, uint8_t transfer_enable, pdcp_device* owner){
	channel->node_id = node_id;
	channel->channel_index = channel_index;
	channel->type = type;
	channel->profile = profile;
	channel->transfer_enable = transfer_enable;
	channel->required_input_channels = 0;
	channel->desired_input_channels = 0;
	channel->owner = owner;
	channel->n_channel_links = 0;
	channel->channel_links = NULL;
	
}

void initialize_pdcp_input_channel (pdcp_input_channel* channel, uint8_t channel_index, uint16_t type, uint16_t profile, uint8_t transfer_enable, pdcp_device* owner){
	channel->channel_index = channel_index;
	channel->type = type;
	channel->profile = profile;
	channel->terminal_profile = 0;
	channel->transfer_enable = transfer_enable;
	channel->required_input_channels = 0;
	channel->desired_input_channels = 0;
	channel->owner = owner;
	channel->channel_link = NULL;
	
}


pdcp_device* l_find_device(pdcp_device* array, size_t size, uint16_t vid, uint16_t pid, uint16_t snum){
	for (int i = 0; i < size; i++){
		if (array[i].vendor_id == vid && array[i].product_id == pid && array[i].serial_num == snum){
			return &array[i];
		}
	}
	return NULL;
} 

pdcp_device* find_present_device(pdcp_system_info* system_info, uint16_t vid, uint16_t pid, uint16_t snum){
	return l_find_device(&system_info->devices[0], system_info->n_devices, vid, pid, snum);
}

pdcp_device* find_missing_device(pdcp_system_info* system_info, uint16_t vid, uint16_t pid, uint16_t snum){
	return l_find_device(&system_info->missing_devices[0], system_info->n_missing_devices, vid, pid, snum);
}


pdcp_input_channel* find_ichan_in_device(pdcp_device* device, uint8_t channel_index){
	for (int i = 0; i < device->n_input_channels; i++){
		if (device->input_channels[i]->channel_index == channel_index){
			return device->input_channels[i];
		}
	}
	return NULL;
}
pdcp_output_channel* find_ochan_in_device(pdcp_device* device, uint8_t channel_index){
	for (int i = 0; i < device->n_output_channels; i++){
		if (device->output_channels[i]->channel_index == channel_index){
			return device->output_channels[i];
		}
	}
	return NULL;
}

void print_system_info_short(pdcp_system_info* system_info){
	printf("n_devices: %d\n", system_info->n_devices);
	printf("n_missing_devices: %d\n", system_info->n_missing_devices);
	printf("n_input_channels: %d\n", system_info->n_input_channels);
	printf("n_output_channels: %d\n", system_info->n_output_channels);
	printf("n_channel_links: %d\n", system_info->n_channel_links);
}

void print_system_info(pdcp_system_info* system_info){
	print_system_info_short(system_info);
	
	for(int i = 0; i < system_info->n_devices; i++){
		printf("\ndevice %d (%X):\n", i, (unsigned)&system_info->devices[i]);
		if (system_info->devices != NULL)
			print_device(system_info, &system_info->devices[i]);
	}
	for(int i = 0; i < system_info->n_missing_devices; i++){
		printf("\nmissing device %d (%X):\n", i, (unsigned)&system_info->missing_devices[i]);
		if (system_info->missing_devices != NULL)
			print_device(system_info, &system_info->missing_devices[i]);
	}
	for(int i = 0; i < system_info->n_input_channels; i++){
		printf("\ninput channel %d (%X):\n", i, (unsigned)&system_info->input_channels[i]);
		if (system_info->input_channels != NULL)
			print_input_channel(system_info, &system_info->input_channels[i]);
	}
	for(int i = 0; i < system_info->n_output_channels; i++){
		printf("\noutput channel %d (%X):\n", i, (unsigned)&system_info->output_channels[i]);
		if (system_info->output_channels != NULL)
			print_output_channel(system_info, &system_info->output_channels[i]);	
	}
	for(int i = 0; i < system_info->n_missing_devices; i++){
		if (system_info->missing_devices != NULL){
			for(int j = 0; j < system_info->missing_devices[i].n_input_channels; j++){
				printf("%d missing input channel %d (%X):\n", i, j, (unsigned)system_info->missing_devices[i].input_channels[j]);
				if (system_info->missing_devices[i].input_channels != NULL)
					print_input_channel(system_info, system_info->missing_devices[i].input_channels[j]);
			}
			for(int j = 0; j < system_info->missing_devices[i].n_output_channels; j++){
				printf("%d missing output channel %d (%X):\n", i, j, (unsigned)system_info->missing_devices[i].output_channels[j]);
				if (system_info->missing_devices[i].output_channels != NULL)
					print_output_channel(system_info, system_info->missing_devices[i].output_channels[j]);	
			}
		}
	}
	for(int i = 0; i < system_info->n_channel_links; i++){
		printf("channel link %d (%X):\n", i, (unsigned)&system_info->channel_links[i]);
		if (system_info->channel_links != NULL)
			print_channel_link(system_info, &system_info->channel_links[i]);
	}
}

void print_device(pdcp_system_info* system_info, pdcp_device* device){
	printf("\tname: %s\n", device->name);
	printf("\tvendor id: %X\n", device->vendor_id);
	printf("\tproduct id: %X\n", device->product_id);
	printf("\tserial number: %X\n", device->serial_num);
	printf("\tnode_id: %d\n", device->node_id);
	printf("\ttype: %X\n", device->type);
	printf("\tprofile: %X\n", device->profile);
	printf("\t# of input channels: %d\n", device->n_input_channels);
	printf("\t# of output channels: %d\n", device->n_output_channels);
	
	printf("\tinput channels: ");
	for (int i = 0; i < device->n_input_channels; i++){
		printf("%X, ", (unsigned)device->input_channels[i]);
		// for (int j = 0; j < system_info->n_input_channels; j++){
			// if(device->input_channels[i] == &system_info->input_channels[j]){
				// printf("%d, ", j);
				// break;
			// }
		// }
	}
	printf("\n");
	
	printf("\toutput channels: ");
	for (int i = 0; i < device->n_output_channels; i++){
		printf("%X, ", (unsigned)device->output_channels[i]);
		// for (int j = 0; j < system_info->n_output_channels; j++){
			// if(device->output_channels[i] == &system_info->output_channels[j]){
				// printf("%d, ", j);
				// break;
			// }
		// }
	}
	printf("\n");
}

void print_input_channel(pdcp_system_info* system_info, pdcp_input_channel* channel){
	printf("\tchannel index: %d\n", channel->channel_index);
	printf("\ttype: %X\n", channel->type);
	printf("\tprofile: %X\n", channel->profile);
	printf("\ttransfer enable: %d\n", channel->transfer_enable);
	printf("\trequired ichans: %X\n", channel->required_input_channels);
	printf("\tdesired ichans: %X\n", channel->desired_input_channels);
	printf("\tterminal profile: %X\n", channel->terminal_profile);
	printf("\tsource channel node id: %d\n", channel->source_channel_node_id);
	printf("\tsource device node id: %d\n", channel->source_device_node_id);
	
	uint8_t source_vid_pid[4];
	uint16_t* source_vid = (uint16_t*)&source_vid_pid[0];
	uint16_t* source_pid = (uint16_t*)&source_vid_pid[2];
	uint8_t source_sn_ci[3];
	uint16_t* source_sn = (uint16_t*)&source_sn_ci[0];
	uint8_t* source_ci = &source_sn_ci[2];
	uint8_t source_type_profile[4];
	uint16_t* source_type = (uint16_t*)&source_type_profile[0];
	uint16_t* source_profile = (uint16_t*)&source_type_profile[2];
	
	get_input_channel_parameter(channel, 5, &source_vid_pid[0], 4);
	get_input_channel_parameter(channel, 6, &source_sn_ci[0], 3);
	get_input_channel_parameter(channel, 8, &source_type_profile[0], 4);
	
	printf("\tsource vid: %X\n", *source_vid);
	printf("\tsource pid: %X\n", *source_pid);
	printf("\tsource snum: %X\n", *source_sn);
	printf("\tsource ci: %X\n", *source_ci);
	printf("\tsource type: %X\n", *source_type);
	printf("\tsource profile: %X\n", *source_profile);
	
	
	printf("\towner: ");
	printf("%X", (unsigned)channel->owner);

	printf("\n");
	
	printf("\tchannel link: ");
	printf("%X", (unsigned)channel->channel_link);

	printf("\n");
	printf("\n");
	
}

void print_output_channel(pdcp_system_info* system_info, pdcp_output_channel* channel){
	printf("\tnode id: %d\n", channel->node_id);
	printf("\tchannel index: %d\n", channel->channel_index);
	printf("\ttype: %X\n", channel->type);
	printf("\tprofile %X\n", channel->profile);
	printf("\tnode id %d\n", channel->node_id);
	printf("\ttransfer enable: %X\n", channel->transfer_enable);
	printf("\trequired ichans: %d\n", channel->required_input_channels);
	printf("\tdesired ichans: %d\n", channel->desired_input_channels);
	
	printf("\towner: ");
	printf("%X\n", (unsigned)channel->owner);

	
	printf("\t# of channel links: %d\n", channel->n_channel_links);
	printf("\tchannel links: ");
	for (int i = 0; i < channel->n_channel_links; i++){
		printf("%X, ", (unsigned)channel->channel_links[i]);

	}
	printf("\n");
	printf("\n");
	
}


void print_channel_link(pdcp_system_info* system_info, pdcp_channel_link* link){
	printf("\tinput channel:");
	printf("%X", (unsigned)link->input_channel);

	printf("\n");
	
	printf("\toutput channel:");
	printf("%X", (unsigned)link->output_channel);

	printf("\n");
	
	printf("\tlink status: %d\n", (unsigned)link->link_status);
	
	printf("\n");
}

int links_equal(pdcp_channel_link* link1, pdcp_channel_link* link2){
	return(	input_channels_equal(link1->input_channel, link2->input_channel) &&
			output_channels_equal(link1->output_channel, link2->output_channel) &&
			link1->link_status == link2->link_status);
}
int input_channels_equal(pdcp_input_channel* chan1, pdcp_input_channel* chan2){
	return(	devices_equal(chan1->owner, chan2->owner) &&
			chan1->channel_index == chan2->channel_index	);
}
int output_channels_equal(pdcp_output_channel* chan1, pdcp_output_channel* chan2){
	return(	devices_equal(chan1->owner, chan2->owner) &&
			chan1->channel_index == chan2->channel_index	);
}
int devices_equal(pdcp_device* dev1, pdcp_device* dev2){
	return(	dev1->vendor_id == dev2->vendor_id &&
		dev1->product_id == dev2->product_id &&
		dev1->serial_num == dev2->serial_num );
}

int system_info_equal(pdcp_system_info* system1, pdcp_system_info* system2){
	int equal;
	equal = system1->n_devices == system2->n_devices;
	equal = equal && (system1->n_input_channels == system2->n_input_channels);
	equal = equal && (system1->n_output_channels == system2->n_output_channels);
	equal = equal && (system1->n_channel_links == system2->n_channel_links);
	equal = equal && (system1->n_missing_devices == system2->n_missing_devices);
	
	if(!equal) return equal;
	
	for (int i = 0; i < system1->n_devices; i++){
		equal = equal && devices_equal(&system1->devices[i], &system2->devices[i]);
	}
	
	if(!equal) return equal;
	
	for (int i = 0; i < system1->n_input_channels; i++){
		equal = equal && input_channels_equal(&system1->input_channels[i], &system2->input_channels[i]);
	}
	
	if(!equal) return equal;
	
	for (int i = 0; i < system1->n_output_channels; i++){
		equal = equal && output_channels_equal(&system1->output_channels[i], &system2->output_channels[i]);
	}
	
	if(!equal) return equal;
	
	for (int i = 0; i < system1->n_channel_links; i++){
		equal = equal && links_equal(&system1->channel_links[i], &system2->channel_links[i]);
	}
	
	if(!equal) return equal;
	
	for (int i = 0; i < system1->n_missing_devices; i++){
		equal = equal && devices_equal(&system1->missing_devices[i], &system2->missing_devices[i]);
	}
	
	return equal;
	
}

int is_missing(pdcp_system_info* system, pdcp_device* device){
	pdcp_device* temp1 = NULL;
	pdcp_device* temp2 = NULL;
	temp1 = find_present_device(system, device->vendor_id, device->product_id, device->serial_num);
	temp2 = find_missing_device(system, device->vendor_id, device->product_id, device->serial_num);
	
	assert((temp1 == NULL) != (temp2 == NULL)); // XOR
	
	return (temp2 != NULL);
}


