
#include <stdint.h>
#include <stdio.h>
#include <assert.h>
#include <time.h>
//#include <stdlib.h>
#include <unistd.h>

#include "pdcp_profiles.h"
#include "pdcp_configurations.h"
#include "pdcp_low_level.h"
#include "pdcp_malloc.h"
#include "pdcp_error.h"
#include "pdcp_names.h"



#define check(ret_code) check_return_code(ret_code, __LINE__)


pdcp_system_info system_info;


int implement_configuration(const pdcp_configuration* conf);
int make_link_from_configuration(pdcp_channel_link_configuration* conf);
void find_replacements();
pdcp_output_channel* find_ochan_replacement(pdcp_output_channel* channel);
pdcp_input_channel* find_ichan_replacement(pdcp_input_channel* channel);
void enforce_sole_access();
int ochan_is_free(pdcp_output_channel* channel);
uint8_t determine_similarity_ichan(pdcp_input_channel* channel1, pdcp_input_channel* channel2);
uint8_t determine_similarity_ochan(pdcp_output_channel* channel1, pdcp_output_channel* channel2);


int is_terminal_channel(pdcp_input_channel* channel);

// returns 1 on success
int intelligently_connect_channel(pdcp_input_channel* channel);
int intelligently_connect_channels(pdcp_input_channel* ichannel, pdcp_output_channel* ochannel);

void refresh_links();
void autoconfigure();
int input_channel_comparator(const void* elem1, const void* elem2);
//int decide_channel_configure_order(pdcp_input_channel** out);
int decide_channel_configure_order(uint16_t** out);

void profiles_init(){

	return_code ret = low_level_init(&system_info);
	check(ret);
	
}

int profiles_ready_to_run_first_setup(){
	// struct timespec tim;// = {.tv_sec = 0, .tv_nsec = PROFILES_TIME_BUFFER_BEFORE_SETUP_US*1000};
	// tim.tv_sec = 0;
	// tim.tv_nsec = 500000000L; //PROFILES_TIME_BUFFER_BEFORE_SETUP_US*1000;
	
	//nanosleep(&tim, NULL);

	usleep(PROFILES_TIME_BUFFER_BEFORE_SETUP_US);
	
	return(refresh_system_info(&system_info));
}

void profiles_first_setup(){
	printf("running first profiles_first_setup\n");
	const pdcp_configuration* loaded_configuration = load_configuration();
	if(loaded_configuration != NULL){
		printf("found configuration\n");
		if (implement_configuration(loaded_configuration))
			printf("configuration implemented successfully\n");
		else
			printf("failed implementing configuration\n");
	} else {
		printf("no configuration loaded\n");
	}
	
	if(system_info.n_missing_devices != 0){
		printf("there are devices missing\n");
		find_replacements();
	} else {
		printf("no devices are missing\n");
	}
	
	printf("autoconfiguring\n");
	autoconfigure(&system_info);
	
	//save_configuration(&system_info);
	
	printf("profiles_first_setup done\n\n");
	
}




int profiles_duty_cycle(){
	int change_has_happened = refresh_system_info(&system_info);
	
	if(change_has_happened){
		if(system_info.n_missing_devices != 0){
			find_replacements();
		}
	
		autoconfigure(&system_info);
		
		refresh_system_info(&system_info);
		
		return 1;
	}
	
	return 0;
}




// Returns 1 if successful, 0 otherwise
int make_link_from_configuration(pdcp_channel_link_configuration* conf){

	pdcp_device* odev = find_present_device(&system_info, conf->output.owner_vid, conf->output.owner_pid, conf->output.owner_snum);
	pdcp_device* idev = find_present_device(&system_info, conf->input.owner_vid,  conf->input.owner_pid,  conf->input.owner_snum);

	int omissing = (odev == NULL);
	int imissing = (idev == NULL);
	
	pdcp_output_channel* ochan = NULL;
	pdcp_input_channel* ichan = NULL;
	
	if (!omissing) ochan = find_ochan_in_device(odev, conf->output.channel_index);
	if (!imissing) ichan = find_ichan_in_device(idev,  conf->input.channel_index);

	omissing = (ochan == NULL);
	imissing = (ichan == NULL);
	
	if (omissing && !imissing){
		printf("finding replacement for output channel\n");
		pdcp_output_channel missing_ochan = {
			.type = conf->output.type,
			.profile = conf->output.profile,
			.owner = NULL
		};
		ochan = find_ochan_replacement(&missing_ochan);
		if (ochan == NULL){
			printf("error: found no replacement");
			return 0;
		}
	} else if (!omissing && imissing){
		printf("finding replacement for input channel\n");
		pdcp_input_channel missing_ichan = {
			.type = conf->input.type,
			.profile = conf->input.profile,
			.owner = NULL
		};
		ichan = find_ichan_replacement(&missing_ichan);
		if (ichan == NULL){
			printf("error: found no replacement");
			return 0;
		}
	} else if (imissing && omissing) {
		printf ("error: both channels are missing, cannot make link from configuration\n");
		return 0;
	}
/*	return_code ret = make_channel_link_raw(	conf->output.owner_vid, conf->output.owner_pid, */
/*							conf->output.owner_snum, conf->output.channel_index, */
/*							conf->input.owner_vid, conf->input.owner_pid, */
/*							conf->input.owner_snum, conf->input.channel_index);*/
/*	*/

	return_code ret = make_channel_link(&system_info, ichan, ochan);
	
	if (ret != OUT_VAR_FILLED){
		printf("error making link from configuration (ret: %d)\n", ret);
		return 0;
	}
	return 1;
}

// Returns 1 if successful, 0 otherwise
int implement_configuration(const pdcp_configuration* conf){
	int ret_value = 1;
	
	for (int i = 0; i < system_info.n_channel_links; i++){
		remove_channel_link(&system_info, &system_info.channel_links[i]);
	}
	
	for (int i = 0; i < conf->n_links; i++){
		ret_value = ret_value && make_link_from_configuration(&conf->links[i]);
	}
	
	refresh_system_info(&system_info);
	
	return ret_value;
}


void find_replacements(){

	
	for (int i = 0; i < system_info.n_channel_links; i++){
		pdcp_channel_link* link = &system_info.channel_links[i];
	
		
		if (link->link_status == LINK_STATUS_INACTIVE){
			int out_missing = is_missing(&system_info, link->output_channel->owner);
			int in_missing = is_missing(&system_info, link->input_channel->owner);
			
			i = -1;
			
			if (out_missing && !in_missing){
				pdcp_output_channel* missing = link->output_channel;
		
/*				uint8_t greatest_similarity = 0;*/
/*				pdcp_output_channel* most_similar;*/
/*		*/
/*				for (int j = 0; j < system_info.n_output_channels; j++){*/
/*					pdcp_output_channel* channel = &system_info.output_channels[j];*/
/*			*/
/*					uint8_t temp_similarity = determine_similarity_ochan(channel, missing);*/
/*			*/
/*					if (temp_similarity > greatest_similarity){*/
/*						greatest_similarity = temp_similarity;*/
/*						most_similar = channel;*/
/*					}*/
/*				}*/
/*				*/
/*				if (greatest_similarity > 1) 	intelligently_connect_channels(link->input_channel, most_similar);//make_channel_link(&system_info, link->input_channel, most_similar);*/

				pdcp_output_channel* replacement = find_ochan_replacement(missing);
				if (replacement != NULL) 	intelligently_connect_channels(link->input_channel, replacement);
				else 				remove_channel_link(&system_info, link);
				
				
			} else if (in_missing && !out_missing){
				pdcp_input_channel* missing = link->input_channel;
		
/*				uint8_t greatest_similarity = 0;*/
/*				pdcp_input_channel* most_similar;*/
/*		*/
/*				for (int j = 0; j < system_info.n_input_channels; j++){*/
/*					pdcp_input_channel* channel = &system_info.input_channels[j];*/
/*			*/
/*					uint8_t temp_similarity = determine_similarity_ichan(channel, missing);*/
/*			*/
/*					if (temp_similarity > greatest_similarity){*/
/*						greatest_similarity = temp_similarity;*/
/*						most_similar = channel;*/
/*					}*/
/*				}*/
/*				*/
				//if (greatest_similarity > 1) 	intelligently_connect_channels(most_similar, link->output_channel);
				
				pdcp_input_channel* replacement = find_ichan_replacement(missing);
				if (replacement != NULL) 	intelligently_connect_channels(replacement, link->output_channel);
				else 				remove_channel_link(&system_info, link);
			}
		}
	}
}





pdcp_output_channel* find_ochan_replacement(pdcp_output_channel* channel){
	uint8_t greatest_similarity = 0;
	pdcp_output_channel* most_similar;

	for (int j = 0; j < system_info.n_output_channels; j++){
		pdcp_output_channel* candidate = &system_info.output_channels[j];

		uint8_t temp_similarity;
		determine_similarity(candidate, channel, temp_similarity);

		if (temp_similarity > greatest_similarity){
			greatest_similarity = temp_similarity;
			most_similar = candidate;
		}
	}
	
	if (greatest_similarity > 1) 	return most_similar;
	else 				return NULL;
}

pdcp_input_channel* find_ichan_replacement(pdcp_input_channel* channel){
	uint8_t greatest_similarity = 0;
	pdcp_input_channel* most_similar;

	for (int j = 0; j < system_info.n_input_channels; j++){
		pdcp_input_channel* candidate = &system_info.input_channels[j];

		uint8_t temp_similarity;
		determine_similarity(candidate, channel, temp_similarity);

		if (temp_similarity > greatest_similarity){
			greatest_similarity = temp_similarity;
			most_similar = candidate;
		}
	}
	
	if (greatest_similarity > 1) 	return most_similar;
	else 				return NULL;
}


void enforce_sole_access(){
	// iterate over output channels
	for (int i = 0; i < system_info.n_output_channels; i++){
		int sole_access = 0;
		
		// iterate over this output channel's links
		for (int j = 0; j < system_info.output_channels[i].n_channel_links; j++){
			if (system_info.output_channels[i].channel_links[j]->input_channel->type & (1 << TYPE_OFFSET_SOLE_ACCESS)){
				// a connected input channel has sole access
				if (!sole_access){
					// first
					sole_access = 1;
				} else {
					// second, sole access has been broken, remove all channel links
					for (int k = 0; k < system_info.n_channel_links; k++){
						remove_channel_link(&system_info, &system_info.channel_links[k]);
					}
				}
			}
		}
	}
}


int ochan_is_free(pdcp_output_channel* channel){
	for (int i = 0; i < channel->n_channel_links; i++){
		// check "sole access" flag in connected input channels
		if(channel->channel_links[i]->input_channel->type & (1 << TYPE_OFFSET_SOLE_ACCESS)) return 0;
	}
	return 1;
}


uint8_t determine_similarity_ichan(pdcp_input_channel* channel1, pdcp_input_channel* channel2){
	uint8_t similarity = 0;
	uint8_t done = 0;
	
	for (int i = 0; i < 4 && !done; i++){
		if ((channel1->profile & (0xF000 >> (i*4))) == (channel2->profile & (0xF000 >> (i*4)))){
			similarity++;
		} else {
			done = 1;
		}
	}
	
	return similarity;
}

uint8_t determine_similarity_ochan(pdcp_output_channel* channel1, pdcp_output_channel* channel2){
	uint8_t similarity = 0;
	uint8_t done = 0;
	
	for (int i = 0; i < 4 && !done; i++){
		if ((channel1->profile & (0xF000 >> (i*4))) == (channel2->profile & (0xF000 >> (i*4)))){
			similarity++;
		} else {
			done = 1;
		}
	}
	
	return similarity;
}

pdcp_output_channel* find_appropriate_channel(pdcp_input_channel* channel){
	uint8_t greatest_similarity = 0;
	pdcp_output_channel* most_similar;

	for (int j = 0; j < system_info.n_output_channels; j++){
		pdcp_output_channel* candidate = &system_info.output_channels[j];

		if ((channel->owner != candidate->owner) && (!(channel->type & (1 << TYPE_OFFSET_SOLE_ACCESS)) || ochan_is_free(candidate))){
			uint8_t temp_similarity;
			determine_similarity(candidate, channel, temp_similarity);

			if (temp_similarity > greatest_similarity){
				greatest_similarity = temp_similarity;
				most_similar = candidate;
			}
		}
	}
	if (greatest_similarity >= 1)
		return most_similar;
	else
		return NULL;
}

int intelligently_connect_channels(pdcp_input_channel* ichannel, pdcp_output_channel* ochannel){

	return_code ret = make_channel_link(&system_info, ichannel, ochannel);

	if (ret == OUT_VAR_FILLED){
		// link was made, find required channels
		int all_ok = 1;
		uint16_t connections_made[16][4];
		connections_made[0][0] = ichannel->owner->vendor_id;
		connections_made[0][1] = ichannel->owner->product_id;
		connections_made[0][2] = ichannel->owner->serial_num;
		connections_made[0][3] = ichannel->channel_index;
		int n_connections_made = 1;
		
		refresh_system_info(&system_info);
		
		pdcp_device* owner = find_present_device(&system_info, connections_made[0][0], connections_made[0][1], connections_made[0][2]);
		ichannel = find_ichan_in_device(owner, connections_made[0][3]);
		ochannel = ichannel->channel_link->output_channel;
		
		printf("owner: %X\n", owner);
		printf("ichannel: %X\n", ichannel);
		printf("ochannel: %X\n", ochannel);
		
		for (uint8_t j = 0; j < 8; j++){
			if (ichannel->required_input_channels & (0x80 >> j)){
				printf("required input channel: %d\n", j);
				pdcp_input_channel* required_ichan = find_ichan_in_device(ichannel->owner, j);
				if (required_ichan == NULL){
					printf("required channel is not present (dev: %X, CI: %d)\n", ichannel->owner, j);
					all_ok = 0;
				} else if (required_ichan->channel_link == NULL){
					if (intelligently_connect_channel(required_ichan)){
						connections_made[n_connections_made][0] = required_ichan->owner->vendor_id;
						connections_made[n_connections_made][1] = required_ichan->owner->product_id;
						connections_made[n_connections_made][2] = required_ichan->owner->serial_num;
						connections_made[n_connections_made][3] = required_ichan->channel_index;
						n_connections_made++;
					} else {
						all_ok = 0;
					}
				}
			}
			if (ochannel->required_input_channels & (0x80 >> j)){
				printf("required input channel: %d\n", j);
				pdcp_input_channel* required_ichan = find_ichan_in_device(ochannel->owner, j);
				if (required_ichan == NULL){
					printf("required channel is not present (dev: %X, CI: %d)\n", ochannel->owner, j);
					all_ok = 0;
				} else if (required_ichan->channel_link == NULL){
					if (intelligently_connect_channel(required_ichan)){
						connections_made[n_connections_made][0] = required_ichan->owner->vendor_id;
						connections_made[n_connections_made][1] = required_ichan->owner->product_id;
						connections_made[n_connections_made][2] = required_ichan->owner->serial_num;
						connections_made[n_connections_made][3] = required_ichan->channel_index;
						n_connections_made++;
					} else {
						all_ok = 0;
					}
				}
			}	
		}
		if (!all_ok){
			for (int j = 0; j < n_connections_made; j++){
				make_channel_link_raw(0xFFFF, 0xFFFF, 0xFFFF, 0xFF, connections_made[j][0], connections_made[j][1], connections_made[j][2], connections_made[j][3]);
			}
				
		}
		return all_ok;
	} else {
		printf("could not connect channel\n");
		return 0;
	}
}


void refresh_links(){
	pdcp_system_info temp;
	return_code ret = get_system_info(&temp);
	check_return_code(ret, __LINE__);
	
	pdcp_free(system_info.channel_links);
	
	system_info.n_channel_links = temp.n_channel_links;
	system_info.channel_links = NULL;
	
	ret = explore_channel_links(&system_info);
	check_return_code(ret, __LINE__);
}

int intelligently_connect_channel(pdcp_input_channel* channel){
		
	pdcp_output_channel* match = find_appropriate_channel(channel);
	
	if (match != NULL){
		printf("match found (profile: %X)\n", channel->profile);
		return intelligently_connect_channels(channel, match);
	} else {
		printf("could not find matching output channel (profile: %X)\n", channel->profile);
		return 0;
	}
}

/*void autoconfigure(){*/
/*	*/
/*	pdcp_input_channel** channel_configure_order = pdcp_malloc(system_info.n_input_channels * sizeof(pdcp_input_channel*));*/
/*	*/
/*	int n_terminal_channels = decide_channel_configure_order(channel_configure_order);*/
/*	*/
/*	uint16_t connections_made[16][4];*/
/*	*/
/*	printf("n_terminal_channels: %d\n", n_terminal_channels);*/
/*	*/
/*	for (int i = 0; i < n_terminal_channels; i++){*/
/*		int connect_ok = intelligently_connect_channel(channel_configure_order[i]);	*/
/*		if (!connect_ok) printf("failed to connect\n");*/
/*	}*/
/*	*/
/*	*/
/*	pdcp_free(channel_configure_order);*/
/*	*/
/*}*/

void autoconfigure(){
	
	uint16_t** channel_configure_order = pdcp_malloc(system_info.n_input_channels * sizeof(uint16_t*));
	for (int i = 0; i < system_info.n_input_channels; i++){
		channel_configure_order[i] = pdcp_malloc(4 * sizeof(uint16_t));
	}
	
	int n_terminal_channels = decide_channel_configure_order(channel_configure_order);
	
	printf("n_terminal_channels: %d\n", n_terminal_channels);
	
	for (int i = 0; i < n_terminal_channels; i++){
		int connect_ok = intelligently_connect_channel(find_ichan_in_device(find_present_device(&system_info, channel_configure_order[i][0], channel_configure_order[i][1], channel_configure_order[i][2]), channel_configure_order[i][3]));	
		if (!connect_ok) printf("failed to connect\n");
	}
	
	for (int i = 0; i < system_info.n_input_channels; i++){
		pdcp_free(channel_configure_order[i]);
	}
	pdcp_free(channel_configure_order);
	
}

int input_channel_comparator(const void* elem1, const void* elem2){
	pdcp_input_channel* channel1 = *(pdcp_input_channel**)elem1;
	pdcp_input_channel* channel2 = *(pdcp_input_channel**)elem2;
	
	     if (channel1->terminal_profile > channel2->terminal_profile) return 1;
	else if (channel1->terminal_profile < channel2->terminal_profile) return -1;
	else if (channel1->profile > channel2->profile) return 1;
	else if (channel1->profile < channel2->profile) return -1;
	
	else return 0;
}

/*int input_channel_comparator(const void* elem1, const void* elem2){*/
/*	uint16_t* channel1 = *(uint16_t**)elem1;*/
/*	uint16_t* channel2 = *(uint16_t**)elem2;*/
/*	*/
/*	     if (channel1->terminal_profile > channel2->terminal_profile) return 1;*/
/*	else if (channel1->terminal_profile < channel2->terminal_profile) return -1;*/
/*	else if (channel1->profile > channel2->profile) return 1;*/
/*	else if (channel1->profile < channel2->profile) return -1;*/
/*	*/
/*	else return 0;*/
/*}*/

/*// returns: number of terminal channels = number of channels in out array*/
/*int decide_channel_configure_order(pdcp_input_channel** out){*/
/*	printf("deciding configure order, n input channels: %d\n", system_info.n_input_channels);*/
/*	*/
/*	int out_it = 0;*/
/*	for(int i = 0; i < system_info.n_input_channels; i++){*/
/*		if (is_terminal_channel(&system_info.input_channels[i]) &&*/
/*		    (system_info.input_channels[i].channel_link == NULL)){*/
/*			*/
/*			out[out_it++] = &system_info.input_channels[i];*/
/*			*/
/*		}*/
/*	}*/
/*	qsort((void*)out, out_it, sizeof(pdcp_input_channel*), input_channel_comparator);*/
/*	*/
/*	return out_it;*/
/*}*/
/*int decide_channel_configure_order(uint16_t** out){*/
/*	printf("deciding configure order, n input channels: %d\n", system_info.n_input_channels);*/
/*	*/
/*	int out_it = 0;*/
/*	for(int i = 0; i < system_info.n_input_channels; i++){*/
/*		if (is_terminal_channel(&system_info.input_channels[i]) &&*/
/*		    (system_info.input_channels[i].channel_link == NULL)){*/
/*			*/
/*			out[out_it++][0] = system_info.input_channels[i].owner->vendor_id;*/
/*			out[out_it++][1] = system_info.input_channels[i].owner->product_id;*/
/*			out[out_it++][2] = system_info.input_channels[i].owner->serial_num;*/
/*			out[out_it++][3] = system_info.input_channels[i].channel_index;*/
/*			*/
/*		}*/
/*	}*/
/*	qsort((void*)out, out_it, sizeof(uint16_t) * 4, input_channel_comparator);*/
/*	*/
/*	return out_it;*/
/*}*/
int decide_channel_configure_order(uint16_t** out){
	printf("deciding configure order, n input channels: %d\n", system_info.n_input_channels);
	
	pdcp_input_channel** channel_configure_order = pdcp_malloc(system_info.n_input_channels * sizeof(pdcp_input_channel*));
	
	int out_it = 0;
	for(int i = 0; i < system_info.n_input_channels; i++){
		if (is_terminal_channel(&system_info.input_channels[i]) &&
		    (system_info.input_channels[i].channel_link == NULL)){
			
			channel_configure_order[out_it++] = &system_info.input_channels[i];
			
		}
	}
	qsort((void*)channel_configure_order, out_it, sizeof(pdcp_input_channel*), input_channel_comparator);
	
	for (int i = 0; i < out_it; i++){
		out[i][0] = channel_configure_order[i]->owner->vendor_id;
		out[i][1] = channel_configure_order[i]->owner->product_id;
		out[i][2] = channel_configure_order[i]->owner->serial_num;
		out[i][3] = channel_configure_order[i]->channel_index;
	}
	
	pdcp_free(channel_configure_order);
	
	return out_it;
}

int is_terminal_channel(pdcp_input_channel* channel){
	return(!!(channel->type & (1 << TYPE_OFFSET_TERMINAL_CHANNEL)));
}







