#include "unity.h"
#include "pdcp_driver.h"
#include "pdcp_names.h"
#include "pdcp_malloc.h"
#include "pdcp_error.h"
#include "fcntl.h"
#include "errno.h"
#include "string.h"
#include "unistd.h"
#include "stdlib.h"


static const uint8_t print_pos = 0;
static const uint8_t print_msgs = 0;
static const uint8_t print_vars = 0;


// dynamically allocates, call driver_free_msg_array() to free. 
uart_msg_array* driver_new_msg_array(const uint16_t payload_size);
void driver_free_msg_array(uart_msg_array* msg);

// when return_code is TOO_SMALL, make out with driver_new_msg_array(in->payload_size);
return_code driver_struct_to_array(const uart_msg_struct* in, uart_msg_array* out);

uint16_t driver_calculate_array_size(const uint16_t payload_size);

// does nothing at this time
uint8_t driver_get_crc8(const uart_msg_struct* in);
// does nothing at this time
uint16_t driver_get_crc16(const uart_msg_struct* in);



uart_msg_array* msg_array;
uart_msg_array* msg_array_w_payload;

uart_conn conn;

uint8_t expected_payload[] = {0x45,0x63,0x54,0x01,0x87};
uart_msg_struct expected_struct = {
	.msg_id = 0xABCD,
	.ref_id = 5,
	.payload_size = 5,
	.payload = &expected_payload[0]
};

uint8_t payload[] = {0x63,0x53,0x01,0x87};
uart_msg_struct msg_struct = {
	.msg_id = 0xBCDE,
	.ref_id = 7,
	.payload_size = 4,
	.payload = &payload[0]
};


uint8_t write_array[] = {0xFE, 0x19, 0xCD, 0xAB, 5, 0, 7, 0, 0x45,0x63,0x54,0x01,0x87,0,0};
uint8_t expected_array[] = {0xFE, 0x19, 0xDE, 0xBC, 7, 0, 6, 0,0x63,0x53,0x01,0x87,0,0};
uint8_t read_array[16] = {0xAB,0xAB,0xAB,0xAB,0xAB,0xAB,0xAB,0xAB,0xAB,0xAB,0xAB,0xAB,0xAB,0xAB,0xAB,0xAB};


void setUp(void){
	
	PRINT_POS(setUp);
	
	memset(read_array, 0xAB, 16*sizeof(uint8_t));
	
	msg_array = driver_new_msg_array(0);
	if (msg_array == NULL){
		TEST_FAIL_MESSAGE("allocation failure\n");
	}
	
	msg_array_w_payload = driver_new_msg_array(14);
	if (msg_array_w_payload == NULL){
		TEST_FAIL_MESSAGE("allocation failure\n");
	}
	
	conn.read_fd = open(READ_PATH, O_RDWR | O_NOCTTY | O_SYNC | O_TRUNC);
	if(conn.read_fd < 0){
		char error_msg[25];
		sprintf(error_msg, "open failed (%s): %d", READ_PATH, errno);
		TEST_FAIL_MESSAGE(error_msg);
	}
	conn.write_fd = conn.read_fd;
	
}

void tearDown(void){
	PRINT_POS(tearDown);
	
	driver_free_msg_array(msg_array);
	driver_free_msg_array(msg_array_w_payload);
	close(conn.read_fd);
	
}

void test_allocateUartMsgArrayNoPayload(){
	uart_msg_array* msg_array;
	
	msg_array = driver_new_msg_array(0);
	
	if (msg_array == NULL){
		TEST_FAIL_MESSAGE("allocation failure\n");
	}
	
	TEST_ASSERT_EQUAL(MIN_PDCP_UART_MSG_SIZE, msg_array->length);
	
	//check for segfault
	msg_array->array[0] = 4;
	
	msg_array->array[msg_array->length-1] = 5;
	

}

void test_allocateUartMsgArrayWithPayload(){
	uart_msg_array* msg_array_w_payload;
	
	msg_array_w_payload = driver_new_msg_array(14);
	
	if (msg_array_w_payload == NULL){
		TEST_FAIL_MESSAGE("allocation failure\n");
	}
	
	TEST_ASSERT_EQUAL(MIN_PDCP_UART_MSG_SIZE + 14 + 2, msg_array_w_payload->length);
	
	//check for segfault
	msg_array_w_payload->array[0] = 4;
	
	msg_array_w_payload->array[msg_array->length-1] = 5;
	

}

void test_structToArrayNoPayload(){
	
	uint8_t expected_array[8] = {0xFE, 0x19, 3, 0, 45, 0, 0, 0};
	uart_msg_struct msg_struct = {
		.msg_id = 3,
		.ref_id = 45,
		.payload_size = 0,
		.payload = NULL
	};
	
	return_code ret = driver_struct_to_array(&msg_struct, msg_array);//, msg_array);

	TEST_ASSERT_EQUAL(OUT_VAR_FILLED, ret);
	TEST_ASSERT_EQUAL_UINT8_ARRAY(expected_array, msg_array->array, 8);
	TEST_ASSERT_EQUAL_UINT8(8, msg_array->length);
}

void test_structToArrayWithPayload(){
	//uart_msg_array* msg_array = driver_new_msg_array(14);
	
	return_code ret = driver_struct_to_array(&msg_struct, msg_array_w_payload);

	TEST_ASSERT_EQUAL(OUT_VAR_FILLED, ret);
	TEST_ASSERT_EQUAL_UINT8(14, msg_array_w_payload->length);
	TEST_ASSERT_EQUAL_HEX8_ARRAY(expected_array, msg_array_w_payload->array, 14);
}


void test_structToArrayFailsWhenArrayIsTooSmall(){
	return_code ret = driver_struct_to_array(&msg_struct, msg_array);
	TEST_ASSERT_EQUAL(TOO_SMALL, ret);
}

void test_arrayToStructFailsWhenOutIsNull(){
	uart_msg_array* msg_array = NULL;
	return_code ret = driver_struct_to_array(&msg_struct, msg_array);
	
	TEST_ASSERT_EQUAL(BAD_POINTER, ret);
}

void test_calculateMsgArraySize(){
	TEST_ASSERT_EQUAL_UINT16(MIN_PDCP_UART_MSG_SIZE, driver_calculate_array_size(0));
	TEST_ASSERT_EQUAL_UINT16(MIN_PDCP_UART_MSG_SIZE + 2 + 35, driver_calculate_array_size(35));
}

void test_getCrc8(){

	TEST_ASSERT_EQUAL_UINT8(0, driver_get_crc8(&expected_struct));
}

void test_getCrc16(){

	TEST_ASSERT_EQUAL_UINT16(0, driver_get_crc16(&expected_struct));
}

void test_sendMsg(){

	ssize_t write_count = driver_send_msg(&conn, &msg_struct);
	
	lseek(conn.read_fd, 0, SEEK_SET);
	
	ssize_t read_count = read(conn.read_fd, &read_array[0], 16*sizeof(uint8_t));
	
	
	TEST_ASSERT_EQUAL_MESSAGE(14, write_count, "write count");
	TEST_ASSERT_EQUAL_MESSAGE(14, read_count, "read count");
	TEST_ASSERT_EQUAL_HEX8_ARRAY(expected_array, read_array, 14);
	TEST_ASSERT_EQUAL_HEX8(0xAB, read_array[14]);
}



void test_newMsgStructNoPayload()
{
	uart_msg_struct* msg;
	
	msg = driver_new_msg_struct(0);
	
	TEST_ASSERT_EQUAL(0, msg->payload_size);
	TEST_ASSERT_NULL(msg->payload);
	
	// check for segfault
	msg->msg_id = 56;
	msg->ref_id = 3;
	msg->payload_size = 3;
	uint8_t payload[3];
	msg->payload = &payload[0];
	
	driver_free_msg_struct(msg);
}

void test_newMsgStructWithPayload()
{
	uart_msg_struct* msg;
	
	msg = driver_new_msg_struct(24);
	
	//printf("msg: %x\nmsg->msg_id: %x\nmsg->ref_id: %x\nmsg->payload_size: %x\nmsg->payload: %x\npayload val: %x\n", msg, &msg->msg_id, &msg->ref_id, &msg->payload_size, &msg->payload, msg->payload);
	
	TEST_ASSERT_EQUAL_UINT16(24, msg->payload_size);
	TEST_ASSERT_NOT_NULL(msg->payload);
	
	// check for segfault
	msg->msg_id = 56;
	msg->ref_id = 3;
	for (int i = 0; i < 24; i++){
		msg->payload[i] = 0xE + i;
	}
	
	TEST_ASSERT_EQUAL_HEX16(56, msg->msg_id);
	TEST_ASSERT_EQUAL_UINT8(3, msg->ref_id);
	TEST_ASSERT_EQUAL_UINT16(24, msg->payload_size);
	TEST_ASSERT_EQUAL_PTR((uint8_t*)msg + sizeof(uart_msg_struct), msg->payload);
	//TEST_ASSERT_EQUAL(sizeof(uart_msg_struct) + 24 - 1, payload[23](uint8_t*)msg)
	msg->payload_size = 3;
	uint8_t payload[3];
	msg->payload = &(payload[0]);
	
	driver_free_msg_struct(msg);
	
}

void test_freeMsgStruct(){
	uart_msg_struct* msg;
	
	int malloc_count_before = get_malloc_count();
	
	msg = driver_new_msg_struct(24);
	if (msg == NULL){
		TEST_FAIL_MESSAGE("allocation failed");
	}
	
	driver_free_msg_struct(msg);
	
	TEST_ASSERT_EQUAL(malloc_count_before, get_malloc_count());
}

void test_freeMsgArray(){
	uart_msg_array* msg;
	
	int malloc_count_before = get_malloc_count();
	
	msg = driver_new_msg_array(24);
	if (msg == NULL){
		TEST_FAIL_MESSAGE("allocation failed");
	}
	
	driver_free_msg_array(msg);
	
	TEST_ASSERT_EQUAL(malloc_count_before, get_malloc_count());
}

void test_checkForMessage(){

	uart_msg_struct* msg;
	uart_msg_struct init_msg = {
		.ref_id = 0x57
	};
	
	init_msg.payload_size = 0;
	init_msg.payload = NULL;

	uint8_t array[60] = {
		0x43, 0x73, 0x95, 0x45, 0x05, 
		0xFE, 0x19, 0x34, 0x12, 0x56, 
		0, 7, 0, 0x45,0x63,
		0x54, 0x01,0x87,0,0,
		0xFF,0xAB,0xAB,0xAB,0xAB,
		0xAB,0xAB,0xAB,0xAB,0xAB,
		0x43, 0x73, 0x95, 0x45, 0x05, 
		0xFE, 0x19, 0x34, 0x12, 0x57, 
		0, 7, 0, 0x45,0x63,
		0x54, 0x01,0x87,0,0,
		0xFF,0xAB,0xAB,0xAB,0xAB,
		0xAB,0xAB,0xAB,0xAB,0xAB
	};
	//driver_send_msg(&conn, &expected_struct);
	
	ssize_t write_count = write(conn.read_fd, array, 60*sizeof(uint8_t));
	check_ret_value((unsigned)write_count, __LINE__);
		
	lseek(conn.read_fd, 0, SEEK_SET);
	
	
	return_code code = driver_check_for_msg(&conn, NULL);
	TEST_ASSERT_EQUAL_MESSAGE(BAD_POINTER, code, "bad pointer");
	// TEST_ASSERT_EQUAL_MESSAGE(5, conn.payload_size_ready, "payload size ready");
	// TEST_ASSERT_EQUAL_MESSAGE(0x56, conn.ref_id_ready, "ref id ready");
	
	code = driver_check_for_msg(&conn, &init_msg);
	TEST_ASSERT_EQUAL_MESSAGE(TOO_SMALL, code, "payload too small");
	TEST_ASSERT_EQUAL_MESSAGE(5, conn.payload_size_ready, "payload size ready");
	// TEST_ASSERT_EQUAL_MESSAGE(0x56, conn.ref_id_ready, "ref id ready");
	
	msg = driver_new_msg_struct(conn.payload_size_ready);
	msg->ref_id = 0x57;
	memset(msg->payload, 0xAB, 5);
	
	code = driver_check_for_msg(&conn, msg);
	TEST_ASSERT_EQUAL_MESSAGE(OUT_VAR_FILLED, code, "message ready");
	
	
	TEST_ASSERT_EQUAL_HEX16(0x1234, msg->msg_id);
	TEST_ASSERT_EQUAL_HEX8(0x57, msg->ref_id);
	TEST_ASSERT_EQUAL_UINT16(5, msg->payload_size);
	TEST_ASSERT_EQUAL_HEX8_ARRAY(expected_payload, msg->payload, 5);
	
	
	code = driver_check_for_msg(&conn, &init_msg);
	TEST_ASSERT_EQUAL_MESSAGE(NO_ACTION, code, "no message");
	
	driver_free_msg_struct(msg);
	
	
}

void test_blockingRead(){
	
	uint8_t payload[5];
	uart_msg_struct msg= {
		.ref_id = 0x57
	};
	msg.payload_size = 5;
	msg.payload = payload;
	
	
	uint8_t array[60] = {
		0x43, 0x73, 0x95, 0x45, 0x05, 
		0xFE, 0x19, 0x34, 0x12, 0x56, 
		0, 7, 0, 0x45,0x63,
		0x54, 0x01,0x87,0,0,
		0xFF,0xAB,0xAB,0xAB,0xAB,
		0xAB,0xAB,0xAB,0xAB,0xAB,
		0x43, 0x73, 0x95, 0x45, 0x05, 
		0xFE, 0x19, 0x34, 0x12, 0x57, 
		0, 7, 0, 0x45,0x63,
		0x54, 0x01,0x87,0,0,
		0xFF,0xAB,0xAB,0xAB,0xAB,
		0xAB,0xAB,0xAB,0xAB,0xAB
	};
	//driver_send_msg(&conn, &expected_struct);
	
	ssize_t write_count = write(conn.read_fd, array, 60*sizeof(uint8_t));
	check_ret_value((unsigned)write_count, __LINE__);
		
	lseek(conn.read_fd, 0, SEEK_SET);
	
	
	return_code ret = driver_blocking_read(&conn, &msg);
	
	TEST_ASSERT_EQUAL(OUT_VAR_FILLED, ret);
	TEST_ASSERT_EQUAL_HEX16(0x1234, msg.msg_id);
	TEST_ASSERT_EQUAL_HEX8(0x57, msg.ref_id);
	TEST_ASSERT_EQUAL_UINT16(5, msg.payload_size);
	TEST_ASSERT_EQUAL_HEX8_ARRAY(expected_payload, msg.payload, 5);
	
	ret = driver_blocking_read(&conn, &msg);
	
	TEST_ASSERT_EQUAL(NO_ACTION, ret);
	
	
}

void test_driverInit(){
	uart_conn conn;
	driver_init(READ_PATH, WRITE_PATH, &conn);
	
	TEST_ASSERT(conn.read_fd >= 0);
	TEST_ASSERT(conn.write_fd >= 0);
}









