#include "unity.h"
#include "pdcp_low_level.h"
#include "pdcp_driver.h"
#include "pdcp_names.h"
#include "pdcp_malloc.h"
#include "pdcp_error.h"
#include "pdcp_test_helper.h"
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>


// dynamically allocates, call driver_free_msg_array() to free. 
uart_msg_array* driver_new_msg_array(const uint16_t payload_size);
void driver_free_msg_array(uart_msg_array* msg);

// when return_code is TOO_SMALL, make out with driver_new_msg_array(in->payload_size);
return_code driver_struct_to_array(const uart_msg_struct* in, uart_msg_array* out);

uint16_t driver_calculate_array_size(const uint16_t payload_size);

// does nothing at this time
uint8_t driver_get_crc8(const uart_msg_struct* in);
// does nothing at this time
uint16_t driver_get_crc16(const uart_msg_struct* in);


extern return_code set_device_parameter(const pdcp_device* dev, const uint8_t parameter_id, const uint8_t* in, const uint16_t in_size);
extern return_code get_device_parameter(const pdcp_device* in, const uint8_t parameter_id, uint8_t* out, const uint16_t out_size);
extern return_code get_input_channel_parameter(const pdcp_input_channel* in, const uint8_t parameter_id, uint8_t* parameter_value, const uint16_t parameter_size);
extern return_code get_output_channel_parameter(const pdcp_output_channel* in, const uint8_t parameter_id, uint8_t* parameter_value, const uint16_t parameter_size);
extern return_code set_input_channel_parameter(const pdcp_input_channel* chan, const uint8_t parameter_id, const uint8_t* parameter_value, const uint16_t parameter_size);
extern return_code set_output_channel_parameter(const pdcp_output_channel* chan, const uint8_t parameter_id, const uint8_t* parameter_value, const uint16_t parameter_size);


uart_conn conn;


void written_bytes_test(uint16_t msg_id, uint16_t payload_size, uint8_t* payload, uint16_t array_size){
/*	uint8_t msg_payload[] = payload; */
	uart_msg_struct test_bytes_msg = { 
		.msg_id = msg_id, 
		.ref_id = 13, 
		.payload_size = payload_size, 
		.payload = payload 
	}; 
	uint8_t test_bytes_array[array_size]; 
	uart_msg_array test_bytes_msg_array = { 
		.length = array_size, 
		.array = &test_bytes_array[0] 
	}; 
	driver_struct_to_array(&test_bytes_msg, &test_bytes_msg_array); 
	uint8_t test_bytes_written_array[array_size]; 
	ssize_t test_bytes_read_count = read(conn.write_fd, &test_bytes_written_array[0], array_size); 
	check_ret_value(test_bytes_read_count, __LINE__); 
	TEST_ASSERT_EQUAL_HEX8_ARRAY(&test_bytes_array[0], &test_bytes_written_array[0], array_size);
}


void setUp(void){
	conn.read_fd = open(READ_PATH, O_RDWR | O_NOCTTY | O_SYNC | O_TRUNC);
	if(conn.read_fd < 0){
		char error_msg[25];
		sprintf(error_msg, "open failed: %d", errno);
	}
	conn.write_fd = open(WRITE_PATH, O_RDWR | O_NOCTTY | O_SYNC | O_TRUNC);
	if(conn.read_fd < 0){
		char error_msg[25];
		sprintf(error_msg, "open failed: %d", errno);
		TEST_FAIL_MESSAGE(error_msg);
	}
	test_helper_init(&conn);
}

void tearDown(void){
	close(conn.read_fd);
}



void test_getSystemInfo(){

	write_failure_response(MSG_ID_GET_SYSTEM_INFO_FAILURE, 4);
	write_system_info_response(12, 34, 56, 78);	
	
	lseek(conn.read_fd, 0, SEEK_SET);
	
	return_code ret = get_system_info(NULL);
	TEST_ASSERT_EQUAL(BAD_POINTER, ret);
	pdcp_system_info system_info;
	ret = get_system_info(&system_info);
	TEST_ASSERT_EQUAL(ERROR_CONDITION, ret);
	TEST_ASSERT_EQUAL(4, pdcp_errno);
	
	ret = get_system_info(&system_info);
	TEST_ASSERT_EQUAL(OUT_VAR_FILLED, ret);
	
	TEST_ASSERT_EQUAL_UINT8(12, system_info.n_devices);
	TEST_ASSERT_EQUAL_UINT8(34, system_info.n_input_channels);
	TEST_ASSERT_EQUAL_UINT8(56, system_info.n_output_channels);
	TEST_ASSERT_EQUAL_UINT8(78, system_info.n_channel_links);
	TEST_ASSERT_NULL(system_info.devices);
	TEST_ASSERT_NULL(system_info.input_channels);
	TEST_ASSERT_NULL(system_info.output_channels);
	TEST_ASSERT_NULL(system_info.channel_links);
	
	
	lseek(conn.write_fd, 0, SEEK_SET);
	written_bytes_test(MSG_ID_GET_SYSTEM_INFO, 0, NULL, 8);
	
}

void test_exploreDevice(){
	
	write_failure_response(MSG_ID_GET_DEVICE_INFO_FAILURE, 5);
	write_device_info_response(0x56, 0x78, 0x89);

	
	lseek(conn.read_fd, 0, SEEK_SET);
	
	return_code ret = explore_device(4, NULL);
	TEST_ASSERT_EQUAL(BAD_POINTER, ret);

	pdcp_device device;
	ret = explore_device(4, &device);
	TEST_ASSERT_EQUAL(ERROR_CONDITION, ret);
	TEST_ASSERT_EQUAL(5, pdcp_errno);
	
	ret = explore_device(4, &device);
	TEST_ASSERT_EQUAL(OUT_VAR_FILLED, ret);
	
	TEST_ASSERT_EQUAL_HEX16(0x1234, device.vendor_id);
	TEST_ASSERT_EQUAL_HEX16(0x2345, device.product_id);
	TEST_ASSERT_EQUAL_HEX16(0x3456, device.serial_num);
	//TEST_ASSERT_EQUAL_HEX8(0x78, device.channel_index);
	TEST_ASSERT_EQUAL_HEX16(0x5678, device.type);
	TEST_ASSERT_EQUAL_HEX16(0x6789, device.profile);
	TEST_ASSERT_EQUAL_HEX8(0, device.n_input_channels);
	TEST_ASSERT_EQUAL_HEX8(0, device.n_output_channels);
	TEST_ASSERT_NULL(device.input_channels);
	TEST_ASSERT_NULL(device.output_channels);
	
	lseek(conn.write_fd, 0, SEEK_SET);
	uint8_t expected_payload[] = {0x04};
	written_bytes_test(MSG_ID_GET_DEVICE_INFO, 1, &expected_payload[0], 11);
	
}

void test_exploreDevices(){

	write_failure_response(MSG_ID_GET_DEVICE_INFO_FAILURE, 6);
	write_device_info_response(0x56, 0x78, 0x89);
	write_device_info_response(0x57, 0x79, 0x8A);
	write_device_info_response(0x58, 0x7A, 0x8B);

	
	lseek(conn.read_fd, 0, SEEK_SET);

	
	//pdcp_device device_array[3];
	pdcp_system_info system_info = {
		.n_devices = 3,
		.n_input_channels = 0,
		.n_output_channels = 0,
		.devices = NULL//device_array
	};
	
	return_code ret = explore_devices(NULL);
	TEST_ASSERT_EQUAL(BAD_POINTER, ret);
	ret = explore_devices(&system_info);
	TEST_ASSERT_EQUAL(ERROR_CONDITION, ret);
	TEST_ASSERT_EQUAL(6, pdcp_errno);
	ret = explore_devices(&system_info);
	
	
	TEST_ASSERT_EQUAL_HEX16(0x3456, system_info.devices[0].serial_num);
	TEST_ASSERT_EQUAL_HEX16(0x3457, system_info.devices[1].serial_num);
	TEST_ASSERT_EQUAL_HEX16(0x3458, system_info.devices[2].serial_num);
	
	TEST_ASSERT_EQUAL_HEX16(0x5678, system_info.devices[0].type);
	TEST_ASSERT_EQUAL_HEX16(0x5679, system_info.devices[1].type);
	TEST_ASSERT_EQUAL_HEX16(0x567A, system_info.devices[2].type);
	
	TEST_ASSERT_EQUAL_HEX16(0x6789, system_info.devices[0].profile);
	TEST_ASSERT_EQUAL_HEX16(0x678A, system_info.devices[1].profile);
	TEST_ASSERT_EQUAL_HEX16(0x678B, system_info.devices[2].profile);
}

void test_exploreInputChannel(){

	write_failure_response(MSG_ID_GET_INPUT_CHANNEL_INFO_FAILURE, 7);
	write_input_channel_info_response(0x56, 0x78, 0x5678, 0x6789);
	
	lseek(conn.read_fd, 0, SEEK_SET);
	
	pdcp_device device_array[2] = {
		{
			.vendor_id = 0x1234,
			.product_id = 0x2345,
			.serial_num = 0x3457,
			.n_input_channels = 0,
			.n_output_channels = 0,
			.input_channels = NULL,
			.output_channels = NULL,
		},
		{
			.vendor_id = 0x1234,
			.product_id = 0x2345,
			.serial_num = 0x3456,
			.n_input_channels = 0,
			.n_output_channels = 0,
			.input_channels = NULL,
			.output_channels = NULL,
		}
	};
	pdcp_system_info system_info = {
		.n_devices = 2,
		.n_input_channels = 1,
		.n_output_channels = 0,
		.devices = device_array
	};
	
	return_code ret = explore_input_channel(&system_info, 5, NULL);
	TEST_ASSERT_EQUAL(BAD_POINTER, ret);
	
	pdcp_input_channel channel;
	
	ret = explore_input_channel(&system_info, 5, &channel);
	TEST_ASSERT_EQUAL(ERROR_CONDITION, ret);
	TEST_ASSERT_EQUAL(7, pdcp_errno);
	
	ret = explore_input_channel(&system_info, 5, &channel);
	TEST_ASSERT_EQUAL(OUT_VAR_FILLED, ret);
	
	
	TEST_ASSERT_EQUAL_PTR(&system_info.devices[1], channel.owner);
	TEST_ASSERT_EQUAL_HEX8(0x78, channel.channel_index);
	TEST_ASSERT_EQUAL_HEX16(0x5678, channel.type);
	TEST_ASSERT_EQUAL_HEX16(0x6789, channel.profile);
	TEST_ASSERT_EQUAL_HEX8(1, system_info.devices[1].n_input_channels);
	TEST_ASSERT_EQUAL_HEX8(0, system_info.devices[0].n_input_channels);
	TEST_ASSERT_NULL(channel.channel_link);
	
	lseek(conn.write_fd, 0, SEEK_SET);
	uint8_t expected_payload[] = {0x05};
	written_bytes_test(MSG_ID_GET_INPUT_CHANNEL_INFO, 1, &expected_payload[0], 11);
}

void test_exploreOutputChannel(){

	write_failure_response(MSG_ID_GET_OUTPUT_CHANNEL_INFO_FAILURE, 8);
	write_output_channel_info_response(0x56, 0x78, 0x5678, 0x6789);

	lseek(conn.read_fd, 0, SEEK_SET);
	
	pdcp_device device_array[2] = {
		{
			.vendor_id = 0x1234,
			.product_id = 0x2345,
			.serial_num = 0x3457,
			.n_input_channels = 0,
			.n_output_channels = 0,
			.input_channels = NULL,
			.output_channels = NULL,
		},
		{
			.vendor_id = 0x1234,
			.product_id = 0x2345,
			.serial_num = 0x3456,
			.n_input_channels = 0,
			.n_output_channels = 0,
			.input_channels = NULL,
			.output_channels = NULL,
		}
	};
	pdcp_system_info system_info = {
		.n_devices = 2,
		.n_input_channels = 0,
		.n_output_channels = 1,
		.devices = device_array
	};
	
	return_code ret = explore_output_channel(&system_info, 6, NULL);
	TEST_ASSERT_EQUAL(BAD_POINTER, ret);
	
	pdcp_output_channel channel;
	ret = explore_output_channel(&system_info, 6, &channel);
	TEST_ASSERT_EQUAL(ERROR_CONDITION, ret);
	TEST_ASSERT_EQUAL(8, pdcp_errno);
	
	ret = explore_output_channel(&system_info, 6, &channel);
	TEST_ASSERT_EQUAL(OUT_VAR_FILLED, ret);
	
	TEST_ASSERT_EQUAL_PTR(&system_info.devices[1], channel.owner);
	TEST_ASSERT_EQUAL_HEX8(0x78, channel.channel_index);
	TEST_ASSERT_EQUAL_HEX16(0x5678, channel.type);
	TEST_ASSERT_EQUAL_HEX16(0x6789, channel.profile);
	TEST_ASSERT_EQUAL_HEX8(1, system_info.devices[1].n_output_channels);
	TEST_ASSERT_EQUAL_HEX8(0, system_info.devices[0].n_output_channels);
	TEST_ASSERT_EQUAL(0, channel.n_channel_links);
	TEST_ASSERT_NULL(channel.channel_links);
	
	
	lseek(conn.write_fd, 0, SEEK_SET);
	uint8_t expected_payload[] = {0x06};
	written_bytes_test(MSG_ID_GET_OUTPUT_CHANNEL_INFO, 1, &expected_payload[0], 11);
}


void test_exploreChannels(){

	write_failure_response(MSG_ID_GET_INPUT_CHANNEL_INFO_FAILURE, 9);
	write_input_channel_info_response(0x56, 0x78, 0x5678, 0x6789);
	write_input_channel_info_response(0x57, 0x78, 0x5679, 0x678A);
	write_output_channel_info_response(0x57, 0x78, 0x567A, 0x678B);
	write_output_channel_info_response(0x57, 0x78, 0x567B, 0x678B);
	write_output_channel_info_response(0x57, 0x78, 0x567C, 0x678D);

	
	lseek(conn.read_fd, 0, SEEK_SET);
	
	pdcp_device device_array[2] = {
		{
			.vendor_id = 0x1234,
			.product_id = 0x2345,
			.serial_num = 0x3456,
			.n_input_channels = 0,
			.n_output_channels = 0,
			.input_channels = NULL,
			.output_channels = NULL,
		},
		{
			.vendor_id = 0x1234,
			.product_id = 0x2345,
			.serial_num = 0x3457,
			.n_input_channels = 0,
			.n_output_channels = 0,
			.input_channels = NULL,
			.output_channels = NULL,
		}
	};
	pdcp_system_info system_info = {
		.n_devices = 2,
		.n_input_channels = 2,
		.n_output_channels = 3,
		.devices = device_array
	};
	
	return_code ret = explore_channels(NULL);
	TEST_ASSERT_EQUAL(BAD_POINTER, ret);
	
	ret = explore_channels(&system_info);
	TEST_ASSERT_EQUAL(ERROR_CONDITION, ret);
	TEST_ASSERT_EQUAL(9, pdcp_errno);
	
	ret = explore_channels(&system_info);
	TEST_ASSERT_EQUAL(OUT_VAR_FILLED, ret);

	
	TEST_ASSERT_EQUAL(1, system_info.devices[0].n_input_channels);
	TEST_ASSERT_EQUAL(1, system_info.devices[1].n_input_channels);
	
	TEST_ASSERT_EQUAL(0, system_info.devices[0].n_output_channels);
	TEST_ASSERT_EQUAL(3, system_info.devices[1].n_output_channels);
	
	TEST_ASSERT_EQUAL_HEX16(0x5678, system_info.input_channels[0].type);
	TEST_ASSERT_EQUAL_HEX16(0x5679, system_info.input_channels[1].type);
	TEST_ASSERT_EQUAL_HEX16(0x567A, system_info.output_channels[0].type);
	TEST_ASSERT_EQUAL_HEX16(0x567B, system_info.output_channels[1].type);
	TEST_ASSERT_EQUAL_HEX16(0x567C, system_info.output_channels[2].type);
	
	TEST_ASSERT_EQUAL_PTR(&system_info.input_channels[0], system_info.devices[0].input_channels[0]);
	TEST_ASSERT_EQUAL_PTR(&system_info.input_channels[1], system_info.devices[1].input_channels[0]);
	TEST_ASSERT_EQUAL_PTR(&system_info.output_channels[0], system_info.devices[1].output_channels[0]);
	TEST_ASSERT_EQUAL_PTR(&system_info.output_channels[1], system_info.devices[1].output_channels[1]);
	TEST_ASSERT_EQUAL_PTR(&system_info.output_channels[2], system_info.devices[1].output_channels[2]);

}

void test_exploreChannelLink(){

	write_failure_response(MSG_ID_GET_CHANNEL_LINK_INFO_FAILURE, 10);
	write_channel_link_info_response(0x56, 1, 0x57, 1, 1);

	lseek(conn.read_fd, 0, SEEK_SET);
	
	pdcp_input_channel inchan = {
		.channel_index = 1
	};
	pdcp_output_channel outchan = {
		.channel_index = 1
	};
	
	#ifdef PRINT
	printf("&inchan: %X\n", &inchan);
	printf("&outchan: %X\n", &outchan);
	#endif
	
	pdcp_input_channel* inchan1;
	pdcp_input_channel* inchan2 = &inchan;
	pdcp_output_channel* outchan1 = &outchan;
	pdcp_output_channel* outchan2;
	
	pdcp_device device_array[2] = {
		{
			.vendor_id = 0x1234,
			.product_id = 0x2345,
			.serial_num = 0x3456,
			.n_input_channels = 1,
			.n_output_channels = 1,
			.input_channels = &inchan1,
			.output_channels = &outchan1,
		},
		{
			.vendor_id = 0x1234,
			.product_id = 0x2345,
			.serial_num = 0x3457,
			.n_input_channels = 1,
			.n_output_channels = 1,
			.input_channels = &inchan2,
			.output_channels = &outchan2,
		}
	};
	pdcp_system_info system_info = {
		.n_devices = 2,
		.n_input_channels = 2,
		.n_output_channels = 2,
		.devices = device_array
	};
	
	
	return_code ret = explore_channel_link(&system_info, 0, NULL);
	TEST_ASSERT_EQUAL(BAD_POINTER, ret);
	
	pdcp_channel_link link;
	
	ret = explore_channel_link(&system_info, 0, &link);
	TEST_ASSERT_EQUAL(ERROR_CONDITION, ret);
	TEST_ASSERT_EQUAL(10, pdcp_errno);
	
	ret = explore_channel_link(&system_info, 0, &link);
	TEST_ASSERT_EQUAL(OUT_VAR_FILLED, ret);
	
	TEST_ASSERT_EQUAL_PTR(inchan2, link.input_channel);
	TEST_ASSERT_EQUAL_PTR(outchan1, link.output_channel);
	TEST_ASSERT_EQUAL(1, outchan1->n_channel_links);
	TEST_ASSERT_EQUAL(1, link.link_status);
	
	
	lseek(conn.write_fd, 0, SEEK_SET);
	uint8_t expected_payload[] = {0x00};
	written_bytes_test(MSG_ID_GET_CHANNEL_LINK_INFO, 1, &expected_payload[0], 11);
	
}

void test_exploreChannelLinks(){

	write_system_info_response(3,5,2,0);
	write_device_info_response(0x56, 0x78, 0x89);
	write_device_info_response(0x57, 0x79, 0x8A);
	write_device_info_response(0x58, 0x7A, 0x8B);
	write_input_channel_info_response(0x56, 2, 0x78, 0x89);
	write_input_channel_info_response(0x57, 1, 0x78, 0x89);
	write_input_channel_info_response(0x58, 1, 0x78, 0x89);
	write_input_channel_info_response(0x58, 2, 0x78, 0x89);
	write_input_channel_info_response(0x58, 3, 0x78, 0x89);
	write_output_channel_info_response(0x56, 1, 0x78, 0x89);
	write_output_channel_info_response(0x57, 2, 0x78, 0x89);
	write_failure_response(MSG_ID_GET_CHANNEL_LINK_INFO_FAILURE, 11);
	write_channel_link_info_response(0x56, 1, 0x57, 1, 1);
	write_channel_link_info_response(0x57, 2, 0x56, 2, 1);
	write_channel_link_info_response(0x56, 1, 0x58, 1, 1);
	write_channel_link_info_response(0x57, 2, 0x58, 2, 1);
	write_channel_link_info_response(0x59, 1, 0x58, 3, 2);
	write_channel_link_info_response(0x56, 1, 0x5A, 1, 2);

	lseek(conn.read_fd, 0, SEEK_SET);
	
	pdcp_system_info system_info;
	
	low_level_init(&system_info);
	
	system_info.n_channel_links = 6;
	
	return_code ret = explore_channel_links(NULL);
	TEST_ASSERT_EQUAL(BAD_POINTER, ret);
	
	ret = explore_channel_links(&system_info);
	TEST_ASSERT_EQUAL(ERROR_CONDITION, ret);
	TEST_ASSERT_EQUAL(11, pdcp_errno);
	
	ret = explore_channel_links(&system_info);
	TEST_ASSERT_EQUAL(OUT_VAR_FILLED, ret);
	
	
	//printf("link1: %X \nlink2: %X\nlink3: %X\n", &system_info.channel_links[0], &system_info.channel_links[1], &system_info.channel_links[2]);
	
	//print_system_info(&system_info);
	
	TEST_ASSERT_EQUAL_PTR(&system_info.channel_links[1], system_info.devices[0].input_channels[0]->channel_link);
	TEST_ASSERT_EQUAL_PTR(&system_info.channel_links[0], system_info.devices[1].input_channels[0]->channel_link);
	TEST_ASSERT_EQUAL_PTR(&system_info.channel_links[0], system_info.devices[0].output_channels[0]->channel_links[0]);
	TEST_ASSERT_EQUAL_PTR(&system_info.channel_links[1], system_info.devices[1].output_channels[0]->channel_links[0]);
	TEST_ASSERT_EQUAL(3, system_info.devices[0].output_channels[0]->n_channel_links);
	TEST_ASSERT_EQUAL(2, system_info.devices[1].output_channels[0]->n_channel_links);
	TEST_ASSERT_EQUAL(2, system_info.n_missing_devices);
	TEST_ASSERT_EQUAL(1, system_info.missing_devices[0].n_output_channels);
	TEST_ASSERT_EQUAL(0, system_info.missing_devices[0].n_input_channels);
	TEST_ASSERT_EQUAL(0, system_info.missing_devices[1].n_output_channels);
	TEST_ASSERT_EQUAL(1, system_info.missing_devices[1].n_input_channels);
	TEST_ASSERT_NULL(system_info.missing_devices[0].input_channels);
	TEST_ASSERT_NULL(system_info.missing_devices[1].output_channels);
	TEST_ASSERT_EQUAL_HEX16(0x2345, system_info.missing_devices[0].product_id);
	TEST_ASSERT_EQUAL_HEX16(0x1234, system_info.missing_devices[0].vendor_id);
	TEST_ASSERT_EQUAL_HEX16(0x3459, system_info.missing_devices[0].serial_num);
	TEST_ASSERT_EQUAL_HEX16(0x345A, system_info.missing_devices[1].serial_num);
	TEST_ASSERT_EQUAL(1, system_info.channel_links[0].link_status);
	TEST_ASSERT_EQUAL(2, system_info.channel_links[4].link_status);
	TEST_ASSERT_EQUAL(2, system_info.channel_links[5].link_status);
	TEST_ASSERT_EQUAL_PTR(&system_info.channel_links[4], system_info.missing_devices[0].output_channels[0]->channel_links[0]);
	TEST_ASSERT_EQUAL_PTR(&system_info.channel_links[5], system_info.missing_devices[1].input_channels[0]->channel_link);
}



void test_lowLevelInit(){
	
	write_system_info_response(3,4,2,4);
	write_device_info_response(0x56, 0x78, 0x89);
	write_device_info_response(0x57, 0x79, 0x8A);
	write_device_info_response(0x58, 0x7A, 0x8B);
	write_input_channel_info_response(0x56, 2, 0x78, 0x89);
	write_input_channel_info_response(0x57, 1, 0x78, 0x89);
	write_input_channel_info_response(0x58, 1, 0x78, 0x89);
	write_input_channel_info_response(0x58, 2, 0x78, 0x89);
	write_output_channel_info_response(0x56, 1, 0x78, 0x89);
	write_output_channel_info_response(0x57, 2, 0x78, 0x89);
	write_channel_link_info_response(0x56, 1, 0x57, 1, 1);
	write_channel_link_info_response(0x56, 1, 0x58, 1, 1);
	write_channel_link_info_response(0x57, 2, 0x56, 2, 1);
	write_channel_link_info_response(0x57, 2, 0x58, 2, 1);
	
	lseek(conn.read_fd, 0, SEEK_SET);
	
	pdcp_system_info system_info;
	low_level_init(&system_info);
	
	TEST_ASSERT_EQUAL_PTR(&system_info.channel_links[0], system_info.devices[1].input_channels[0]->channel_link);
	TEST_ASSERT_EQUAL_PTR(&system_info.channel_links[1], system_info.devices[2].input_channels[0]->channel_link);
	TEST_ASSERT_EQUAL_PTR(&system_info.channel_links[2], system_info.devices[0].input_channels[0]->channel_link);
	TEST_ASSERT_EQUAL_PTR(&system_info.channel_links[3], system_info.devices[2].input_channels[1]->channel_link);
	TEST_ASSERT_EQUAL_PTR(&system_info.channel_links[0], system_info.devices[0].output_channels[0]->channel_links[0]);
	TEST_ASSERT_EQUAL_PTR(&system_info.channel_links[1], system_info.devices[0].output_channels[0]->channel_links[1]);
	TEST_ASSERT_EQUAL_PTR(&system_info.channel_links[2], system_info.devices[1].output_channels[0]->channel_links[0]);
	TEST_ASSERT_EQUAL_PTR(&system_info.channel_links[3], system_info.devices[1].output_channels[0]->channel_links[1]);
	TEST_ASSERT_EQUAL(2, system_info.devices[0].output_channels[0]->n_channel_links);
	TEST_ASSERT_EQUAL(2, system_info.devices[1].output_channels[0]->n_channel_links);
	
	TEST_ASSERT_EQUAL_PTR(&system_info.devices[0], system_info.devices[0].output_channels[0]->owner);
	TEST_ASSERT_EQUAL_PTR(&system_info.devices[1], system_info.devices[1].output_channels[0]->owner);
	
}

void test_cleanUpSystemInfo(){
	
	write_system_info_response(3,5,2,5);
	write_device_info_response(0x56, 0x78, 0x89);
	write_device_info_response(0x57, 0x79, 0x8A);
	write_device_info_response(0x58, 0x7A, 0x8B);
	write_input_channel_info_response(0x56, 2, 0x78, 0x89);
	write_input_channel_info_response(0x57, 1, 0x78, 0x89);
	write_input_channel_info_response(0x58, 1, 0x78, 0x89);
	write_input_channel_info_response(0x58, 2, 0x78, 0x89);
	write_input_channel_info_response(0x58, 3, 0x78, 0x89);
	write_output_channel_info_response(0x56, 1, 0x78, 0x89);
	write_output_channel_info_response(0x57, 2, 0x78, 0x89);
	write_channel_link_info_response(0x56, 1, 0x57, 1, 1);
	write_channel_link_info_response(0x56, 1, 0x58, 1, 1);
	write_channel_link_info_response(0x57, 2, 0x56, 2, 1);
	write_channel_link_info_response(0x57, 2, 0x58, 2, 1);
	write_channel_link_info_response(0x59, 1, 0x58, 3, 1);
	
	lseek(conn.read_fd, 0, SEEK_SET);
	
	int malloc_count_before = get_malloc_count();
	
	pdcp_system_info system_info;
	low_level_init(&system_info);
	
	clean_up_system_info(&system_info);
	
	
	TEST_ASSERT_EQUAL(malloc_count_before, get_malloc_count());
}

void test_getDeviceParameter(){
	write_system_info_response(1,0,0,0);
	write_device_info_response(0x56, 0x78, 0x89);
	write_get_parameter_response(6);
	
	lseek(conn.read_fd, 0, SEEK_SET);
	
	pdcp_system_info system_info;
	low_level_init(&system_info);
	
	return_code ret = get_device_parameter(&system_info.devices[0], 0, NULL, 0);
	
	//TEST_ASSERT_EQUAL(BAD_POINTER, ret);
	
	uint8_t* parameter_value = (uint8_t*)pdcp_malloc(conn.payload_size_ready * sizeof(uint8_t));
	
	ret = get_device_parameter(&system_info.devices[0], 0, parameter_value, conn.payload_size_ready);
	
	TEST_ASSERT_EQUAL(OUT_VAR_FILLED, ret);
	
	uint8_t expected_value[6] = {0xA0, 0xA1, 0xA2, 0xA3, 0xA4, 0xA5};
	
	TEST_ASSERT_EQUAL_UINT8_ARRAY(expected_value, parameter_value, 6);
	
	lseek(conn.write_fd, -18, SEEK_CUR);
	uint8_t expected_payload[] = {0x34, 0x12, 0x45, 0x23, 0x56, 0x34, 0x00, 0x00};
	written_bytes_test(MSG_ID_GET_DEVICE_PARAMETER, 8, &expected_payload[0], 18);
	
}

void test_setDeviceParameter(){
	write_system_info_response(1,0,0,0);
	write_device_info_response(0x56, 0x78, 0x89);
	write_set_parameter_response();
	
	lseek(conn.read_fd, 0, SEEK_SET);
	
	pdcp_system_info system_info;
	low_level_init(&system_info);
	
	uint8_t parameter_value[] = {0x12, 0x34, 0x56, 0x78, 0x9A};

	lseek(conn.write_fd, 0, SEEK_SET);
	
	set_device_parameter(&system_info.devices[0], 0, parameter_value, 5);
	
	lseek(conn.write_fd, 0, SEEK_SET);
	uint8_t expected_payload[] = {0x34, 0x12, 0x45, 0x23, 0x56, 0x34, 0x00, 0x00, 0x12, 0x34, 0x56, 0x78, 0x9A};
	written_bytes_test(MSG_ID_SET_DEVICE_PARAMETER, 13, &expected_payload[0], 23);
	
	
}

void test_getChannelParameter(){


	write_system_info_response(1,1,1,0);
	write_device_info_response(0x56, 0x78, 0x89);
	write_input_channel_info_response(0x56, 1, 0x43, 0x21);
	write_output_channel_info_response(0x56, 2, 0x54, 0x32);
	write_get_parameter_response(6);
	write_get_parameter_response(7);
	
	lseek(conn.read_fd, 0, SEEK_SET);
	
	pdcp_system_info system_info;
	low_level_init(&system_info);
	
	
	
	return_code ret = get_input_channel_parameter(system_info.devices[0].input_channels[0], 0, NULL, 0);
	TEST_ASSERT_EQUAL(6, conn.payload_size_ready);
	
	uint8_t parameter_value[7];// = (uint8_t*)pdcp_malloc(conn.payload_size_ready * sizeof(uint8_t));
	ret = get_input_channel_parameter(system_info.devices[0].input_channels[0], 0, parameter_value, 7);
	
	
	
	uint8_t expected_value1[6] = {0xA0, 0xA1, 0xA2, 0xA3, 0xA4, 0xA5};
	TEST_ASSERT_EQUAL_UINT8_ARRAY(expected_value1, parameter_value, 6);
	
	
	ret = get_output_channel_parameter(system_info.devices[0].output_channels[0], 0, NULL, 0);
	TEST_ASSERT_EQUAL(7, conn.payload_size_ready);
	
	
	
	ret = get_output_channel_parameter(system_info.devices[0].output_channels[0], 0, parameter_value, 7);
	
	TEST_ASSERT_EQUAL(OUT_VAR_FILLED, ret);
	
	uint8_t expected_value2[7] = {0xA0, 0xA1, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6};
	
	TEST_ASSERT_EQUAL_UINT8_ARRAY(expected_value2, parameter_value, 7);
	
	lseek(conn.write_fd, -54, SEEK_CUR);
	uint8_t expected_payload[] = {0x34, 0x12, 0x45, 0x23, 0x56, 0x34, 0x01, 0x00};
	written_bytes_test(MSG_ID_GET_DEVICE_PARAMETER, 8, &expected_payload[0], 18);
	
	expected_payload[6] = 2;
	written_bytes_test(MSG_ID_GET_DEVICE_PARAMETER, 8, &expected_payload[0], 18);
	written_bytes_test(MSG_ID_GET_DEVICE_PARAMETER, 8, &expected_payload[0], 18);
	
}

void test_setChannelParameter(){

	write_system_info_response(1,1,1,0);
	write_device_info_response(0x56, 0x78, 0x89);
	write_input_channel_info_response(0x56, 1, 0x43, 0x21);
	write_output_channel_info_response(0x56, 2, 0x54, 0x32);
	write_set_parameter_response();
	write_set_parameter_response();
	
	lseek(conn.read_fd, 0, SEEK_SET);
	
	pdcp_system_info system_info;
	low_level_init(&system_info);
	
	
	uint8_t parameter_value[] = {0x12, 0x34, 0x56, 0x78, 0x9A};

	
	lseek(conn.write_fd, 0, SEEK_SET);
		
	set_input_channel_parameter(system_info.devices[0].input_channels[0], 0, parameter_value, 5);
	set_output_channel_parameter(system_info.devices[0].output_channels[0], 0, parameter_value, 5);
	
	lseek(conn.write_fd, 0, SEEK_SET);
	
	uint8_t expected_payload[] = {0x34, 0x12, 0x45, 0x23, 0x56, 0x34, 0x01, 0x00, 0x12, 0x34, 0x56, 0x78, 0x9A};
	written_bytes_test(MSG_ID_SET_DEVICE_PARAMETER, 13, &expected_payload[0], 23);
	
	expected_payload[6] = 2;
	written_bytes_test(MSG_ID_SET_DEVICE_PARAMETER, 13, &expected_payload[0], 23);
	
	
}

void test_configureChannelLink(){
	
	write_system_info_response(2,1,1,0);
	write_device_info_response(0x56, 0x78, 0x89);
	write_device_info_response(0x57, 0x78, 0x89);
	write_input_channel_info_response(0x56, 1, 0x43, 0x21);
	write_output_channel_info_response(0x57, 1, 0x54, 0x32);
	write_failure_response(MSG_ID_CONFIGURE_CHANNEL_LINK_FAILURE, 13);
	write_configure_channel_link_response(3);

	write_system_info_response(2,1,1,1);
	write_device_info_response(0x56, 0x78, 0x89);
	write_device_info_response(0x57, 0x78, 0x89);
	write_input_channel_info_response(0x56, 1, 0x43, 0x21);
	write_output_channel_info_response(0x57, 1, 0x54, 0x32);
	write_channel_link_info_response(0x57, 1, 0x56, 1, LINK_STATUS_ACTIVE);
	
	lseek(conn.read_fd, 0, SEEK_SET);
	
	pdcp_system_info system_info;
	low_level_init(&system_info);
	
	return_code ret = make_channel_link(&system_info, NULL, NULL);
	TEST_ASSERT_EQUAL(BAD_POINTER, ret);
	ret = make_channel_link(&system_info, NULL, &system_info.output_channels[0]);
	TEST_ASSERT_EQUAL(BAD_POINTER, ret);
	ret = make_channel_link(&system_info, &system_info.input_channels[0], NULL);
	TEST_ASSERT_EQUAL(BAD_POINTER, ret);
	
	lseek(conn.write_fd, 0, SEEK_SET);
	ret = make_channel_link(&system_info, &system_info.input_channels[0], &system_info.output_channels[0]);
	TEST_ASSERT_EQUAL(ERROR_CONDITION, ret);
	TEST_ASSERT_EQUAL(13, pdcp_errno);
	lseek(conn.write_fd, 0, SEEK_SET);
	
	uint8_t expected_payload[] = {0x34, 0x12, 0x45, 0x23, 0x57, 0x34, 0x01, 0x34, 0x12, 0x45, 0x23, 0x56, 0x34, 0x01};
	written_bytes_test(MSG_ID_CONFIGURE_CHANNEL_LINK, 14, &expected_payload[0], 24);
	
	ret = make_channel_link(&system_info, &system_info.input_channels[0], &system_info.output_channels[0]);
	TEST_ASSERT_EQUAL(OUT_VAR_FILLED, ret);
	
	refresh_system_info(&system_info);
	
	TEST_ASSERT_EQUAL(1, system_info.n_channel_links);
	TEST_ASSERT_EQUAL(1, system_info.output_channels[0].n_channel_links);
	TEST_ASSERT_EQUAL_PTR(&system_info.channel_links[0], system_info.input_channels[0].channel_link);
	TEST_ASSERT_EQUAL_PTR(&system_info.channel_links[0], system_info.output_channels[0].channel_links[0]);
	TEST_ASSERT_EQUAL_PTR(&system_info.input_channels[0], system_info.channel_links[0].input_channel);
	TEST_ASSERT_EQUAL_PTR(&system_info.output_channels[0], system_info.channel_links[0].output_channel);
	TEST_ASSERT_EQUAL(LINK_STATUS_ACTIVE, system_info.channel_links[0].link_status);
}

void test_passthroughCommand(){
	
	write_system_info_response(1, 0, 0, 0);
	write_device_info_response(0x56, 0x00, 0x00);
	write_failure_response(MSG_ID_PASSTHROUGH_FAILURE, 15);
	write_failure_response(MSG_ID_PASSTHROUGH_SUCCESS, 15);
/*	write_passthrough_response();*/

	lseek(conn.read_fd, 0, SEEK_SET);
	
	pdcp_system_info system_info;
	low_level_init(&system_info);

	uint8_t can_data[5] = {0xC2, 0xC3, 0xC4, 0xC5, 0xC6};
	return_code ret = passthrough_command(&system_info.devices[0], 5, &can_data[0]);
	TEST_ASSERT_EQUAL(ERROR_CONDITION, ret);
	TEST_ASSERT_EQUAL(15, pdcp_errno);
	
	ret = passthrough_command(&system_info.devices[0], 5, &can_data[0]);
	TEST_ASSERT_EQUAL(OUT_VAR_FILLED, ret);
	
	lseek(conn.write_fd, -25, SEEK_CUR);
	uint8_t expected_payload[] = {0x34, 0x12, 0x45, 0x23, 0x56, 0x34, 5, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0, 0, 0, 0, 0};
	written_bytes_test(MSG_ID_PASSTHROUGH, 15, &expected_payload[0], 25);
}

void test_findDevice(){
	write_system_info_response(2, 1, 1, 0);
	write_device_info_response(0x56, 0x01, 0x02);
	write_device_info_response(0x57, 0x03, 0x04);
	write_input_channel_info_response(0x56, 1, 0x43, 0x21);
	write_output_channel_info_response(0x57, 1, 0x54, 0x32);

	lseek(conn.read_fd, 0, SEEK_SET);
	
	pdcp_system_info system;
	low_level_init(&system);
	
	system.n_devices = 1;
	system.n_missing_devices = 1;
	system.missing_devices = &system.devices[1];
	system.n_output_channels = 0;
	system.output_channels = NULL;
	
	TEST_ASSERT_NULL(find_present_device(&system, 0x1234, 0x2345, 0x3457));
	TEST_ASSERT_NULL(find_missing_device(&system, 0x1234, 0x2345, 0x3456));
	
	TEST_ASSERT_EQUAL_PTR(&system.devices[0], find_present_device(&system, 0x1234, 0x2345, 0x3456));
	TEST_ASSERT_EQUAL_PTR(&system.missing_devices[0], find_missing_device(&system, 0x1234, 0x2345, 0x3457));
	
}

void test_findChannel(){
	write_system_info_response(2, 1, 1, 0);
	write_device_info_response(0x56, 0x01, 0x02);
	write_device_info_response(0x57, 0x03, 0x04);
	write_input_channel_info_response(0x56, 1, 0x43, 0x21);
	write_output_channel_info_response(0x57, 1, 0x54, 0x32);

	lseek(conn.read_fd, 0, SEEK_SET);
	
	pdcp_system_info system;
	low_level_init(&system);
	
	
	TEST_ASSERT_EQUAL_PTR(&system.input_channels[0], find_ichan_in_device(&system.devices[0], 1));
	TEST_ASSERT_EQUAL_PTR(&system.output_channels[0], find_ochan_in_device(&system.devices[1], 1));
	TEST_ASSERT_NULL(find_ichan_in_device(&system.devices[0], 2));
	TEST_ASSERT_NULL(find_ochan_in_device(&system.devices[0], 1));
}
